from dataclasses import dataclass

from qibo import gates
from qibo.config import log, raise_error

from qibolab.transpilers.abstract import Transpiler


def find_connected_qubit(qubits, queue, hardware_qubits):
    """Helper method for :meth:`qibolab.transpilers.fix_connecivity`.

    Finds which qubit should be mapped to hardware middle qubit
    by looking at the two-qubit gates that follow.
    """
    possible_qubits = set(qubits)
    for next_gate in queue:
        if len(next_gate.qubits) == 2:
            possible_qubits &= {hardware_qubits.index(q) for q in next_gate.qubits}
            if not possible_qubits:
                # freedom of choice
                return qubits[0]
            elif len(possible_qubits) == 1:
                return possible_qubits.pop()
    # freedom of choice
    return qubits[0]


@dataclass
class StarConnectivity(Transpiler):
    """Transforms an arbitrary circuit to one that can be executed on hardware.

    This transpiler produces a circuit that respects the following connectivity:


    .. code-block:: text

             q
             |
        q -- q -- q
             |
             q

    by adding SWAP gates when needed. It does not translate gates to native.
    """

    middle_qubit: int
    """Qubit id of the qubit that is in the middle of the star."""
    verbose: bool = False

    def tlog(self, message):
        """Print messages only if ``verbose`` was set to ``True``."""
        if self.verbose:
            log.info(message)

    def is_satisfied(self, circuit, middle_qubit=2, verbose=True):
        """Checks if a circuit respects connectivity constraints.

        Args:
            circuit (qibo.models.Circuit): Circuit model to check.
            middle_qubit (int): Hardware middle qubit.
            verbose (bool): If ``True`` it prints debugging log messages.

        Returns ``True`` if the following conditions are satisfied:
            - Circuit does not contain more than two-qubit gates.
            - All two-qubit gates have qubit 0 as target or control.

            otherwise returns ``False``.
        """
        for gate in circuit.queue:
            if len(gate.qubits) > 2 and not isinstance(gate, gates.M):
                self.tlog(f"{gate.name} acts on more than two qubits.")
                return False
            elif len(gate.qubits) == 2:
                if self.middle_qubit not in gate.qubits:
                    self.tlog("Circuit does not respect connectivity. " f"{gate.name} acts on {gate.qubits}.")
                    return False

        self.tlog("Circuit respects connectivity.")
        return True

    def __call__(self, circuit):
        """Apply the transpiler transformation on a given circuit.

        Args:
            circuit (qibo.models.Circuit): The original Qibo circuit to transform.
                This circuit must contain up to two-qubit gates.

        Returns:
            new (qibo.models.Circuit): Qibo circuit that performs the same operation
                as the original but respects the hardware connectivity.
            hardware_qubits (list): List that maps logical to hardware qubits.
                This is required for transforming final measurements.
        """
        # TODO: Change this to a more lightweight form that takes a list of pairs
        # instead of the whole circuit.

        middle_qubit = self.middle_qubit
        # find the number of qubits for hardware circuit
        if circuit.nqubits == 1:
            nqubits = 1
        else:
            nqubits = max(circuit.nqubits, middle_qubit + 1)
        # new circuit object that will be compatible with hardware connectivity
        new = circuit.__class__(nqubits)
        # list to maps logical to hardware qubits
        hardware_qubits = list(range(nqubits))

        # find initial qubit mapping
        for i, gate in enumerate(circuit.queue):
            if len(gate.qubits) == 2:
                if middle_qubit not in gate.qubits:
                    new_middle = find_connected_qubit(gate.qubits, circuit.queue[i + 1 :], hardware_qubits)
                    hardware_qubits[middle_qubit], hardware_qubits[new_middle] = (
                        hardware_qubits[new_middle],
                        hardware_qubits[middle_qubit],
                    )
                break

        # the first SWAP is not needed as it can be applied via virtual mapping
        add_swap = False
        for i, gate in enumerate(circuit.queue):
            # map gate qubits to hardware
            qubits = tuple(hardware_qubits.index(q) for q in gate.qubits)
            if isinstance(gate, gates.M):
                new_gate = gates.M(*qubits, **gate.init_kwargs)
                new_gate.result = gate.result
                new.add(new_gate)
                continue

            if len(qubits) > 2:
                raise_error(
                    NotImplementedError,
                    "Transpiler does not support gates targeting more than two-qubits.",
                )

            elif len(qubits) == 2 and middle_qubit not in qubits:
                # find which qubit should be moved to 0
                new_middle = find_connected_qubit(qubits, circuit.queue[i + 1 :], hardware_qubits)
                # update hardware qubits according to the swap
                hardware_qubits[middle_qubit], hardware_qubits[new_middle] = (
                    hardware_qubits[new_middle],
                    hardware_qubits[middle_qubit],
                )
                if add_swap:
                    new.add(gates.SWAP(middle_qubit, new_middle))
                # update gate qubits according to the new swap
                qubits = tuple(hardware_qubits.index(q) for q in gate.qubits)

            # add gate to the hardware circuit
            if isinstance(gate, gates.Unitary):
                # gates.Unitary requires matrix as first argument
                from qibo.backends import NumpyBackend

                backend = NumpyBackend()
                matrix = gate.matrix(backend)
                new.add(gate.__class__(matrix, *qubits, **gate.init_kwargs))
            else:
                new.add(gate.__class__(*qubits, **gate.init_kwargs))
            if len(qubits) == 2:
                add_swap = True

        return new, hardware_qubits



from dataclasses import dataclass, field
from typing import Dict, Optional

from qibo.config import raise_error

from qibolab.instruments.oscillator import LocalOscillator
from qibolab.instruments.port import Port


def check_max_offset(offset, max_offset):
    """Checks if a given offset value exceeds the maximum supported offset.

    This is to avoid sending high currents that could damage lab equipment
    such as amplifiers.
    """
    if max_offset is not None and abs(offset) > max_offset:
        raise_error(ValueError, f"{offset} exceeds the maximum allowed offset {max_offset}.")


@dataclass
class Channel:
    """Representation of physical wire connection (channel)."""

    name: str
    """Name of the channel from the lab schematics."""
    port: Optional[Port] = None
    """Instrument port that is connected to this channel."""
    local_oscillator: Optional[LocalOscillator] = None
    """Instrument object controlling the local oscillator connected to this channel.

    Not applicable for setups that do not use external local oscillators because the
    controller can send sufficiently high frequencies or contains internal local
    oscillators.
    """
    max_offset: Optional[float] = None
    """Maximum DC voltage that we can safely send through this channel.

    Sending high voltages for prolonged times may damage amplifiers or other lab equipment.
    If the user attempts to send a higher value an error will be raised to prevent
    execution in real instruments.
    """

    @property
    def offset(self):
        """DC offset that is applied to this port."""
        return self.port.offset

    @offset.setter
    def offset(self, value):
        check_max_offset(value, self.max_offset)
        self.port.offset = value

    @property
    def lo_frequency(self):
        if self.local_oscillator is not None:
            return self.local_oscillator.frequency
        return self.port.lo_frequency

    @lo_frequency.setter
    def lo_frequency(self, value):
        if self.local_oscillator is not None:
            self.local_oscillator.frequency = value
        else:
            self.port.lo_frequency = value

    @property
    def lo_power(self):
        if self.local_oscillator is not None:
            return self.local_oscillator.power
        return self.port.lo_power

    @lo_power.setter
    def lo_power(self, value):
        if self.local_oscillator is not None:
            self.local_oscillator.power = value
        else:
            self.port.lo_power = value

    # TODO: gain, attenuation and power range can be unified to a single property
    @property
    def gain(self):
        return self.port.gain

    @gain.setter
    def gain(self, value):
        self.port.gain = value

    @property
    def attenuation(self):
        return self.port.attenuation

    @attenuation.setter
    def attenuation(self, value):
        self.port.attenuation = value

    @property
    def power_range(self):
        return self.port.power_range

    @power_range.setter
    def power_range(self, value):
        self.port.power_range = value

    @property
    def filters(self):
        return self.port.filters

    @filters.setter
    def filters(self, value):
        self.port.filters = value


@dataclass
class ChannelMap:
    """Collection of :class:`qibolab.designs.channel.Channel` objects identified by name.

    Essentially, it allows creating a mapping of names to channels just
    specifying the names.

    """

    _channels: Dict[str, Channel] = field(default_factory=dict)

    def add(self, *items):
        """Add multiple items to the channel map.

        If :class:`qibolab.channels.Channel` objects are given they are added to the channel map.
        If a different type is given, a :class:`qibolab.channels.Channel` with the
        corresponding name is created and added to the channel map.
        """
        for item in items:
            if isinstance(item, Channel):
                self[item.name] = item
            else:
                self[item] = Channel(item)
        return self

    def __getitem__(self, name):
        return self._channels[name]

    def __setitem__(self, name, channel):
        if not isinstance(channel, Channel):
            raise_error(TypeError, f"Cannot add channel of type {type(channel)} to ChannelMap.")
        self._channels[name] = channel

    def __contains__(self, name):
        return name in self._channels

    def __or__(self, channel_map):
        channels = self._channels.copy()
        channels.update(channel_map._channels)
        return self.__class__(channels)

    def __ior__(self, items):
        if not isinstance(items, type(self)):
            try:
                if isinstance(items, str):
                    raise TypeError
                items = type(self)().add(*items)
            except TypeError:
                items = type(self)().add(items)
        self._channels.update(items._channels)
        return self



import numpy as np
from qblox_instruments.qcodes_drivers.sequencer import Sequencer as QbloxSequencer

from qibolab.instruments.qblox.q1asm import Program
from qibolab.pulses import Pulse, PulseSequence, PulseType
from qibolab.sweeper import Parameter, Sweeper


class WaveformsBuffer:
    """A class to represent a buffer that holds the unique waveforms used by a sequencer.

    Attributes:
        unique_waveforms (list): A list of unique Waveform objects.
        available_memory (int): The amount of memory available expressed in numbers of samples.
    """

    SIZE: int = 16383

    class NotEnoughMemory(Exception):
        """An error raised when there is not enough memory left to add more waveforms."""

    class NotEnoughMemoryForBaking(Exception):
        """An error raised when there is not enough memory left to bake pulses."""

    def __init__(self):
        """Initialises the buffer with an empty list of unique waveforms."""
        self.unique_waveforms: list = []  # Waveform
        self.available_memory: int = WaveformsBuffer.SIZE

    def add_waveforms(self, pulse: Pulse, hardware_mod_en: bool, sweepers: list[Sweeper]):
        """Adds a pair of i and q waveforms to the list of unique waveforms.

        Waveforms are added to the list if they were not there before.
        Each of the waveforms (i and q) is processed individually.

        Args:
            waveform_i (Waveform): A Waveform object containing the samples of the real component of the pulse wave.
            waveform_q (Waveform): A Waveform object containing the samples of the imaginary component of the pulse wave.

        Raises:
            NotEnoughMemory: If the memory needed to store the waveforms in more than the memory avalible.
        """
        pulse_copy = pulse.copy()
        for sweeper in sweepers:
            if sweeper.pulses and sweeper.parameter == Parameter.amplitude:
                if pulse in sweeper.pulses:
                    pulse_copy.amplitude = 1

        baking_required = False
        for sweeper in sweepers:
            if sweeper.pulses and sweeper.parameter == Parameter.duration:
                if pulse in sweeper.pulses:
                    baking_required = True
                    values = sweeper.get_values(pulse.duration)

        if not baking_required:
            if hardware_mod_en:
                waveform_i, waveform_q = pulse_copy.envelope_waveforms
            else:
                waveform_i, waveform_q = pulse_copy.modulated_waveforms

            pulse.waveform_i = waveform_i
            pulse.waveform_q = waveform_q

            if waveform_i not in self.unique_waveforms or waveform_q not in self.unique_waveforms:
                memory_needed = 0
                if not waveform_i in self.unique_waveforms:
                    memory_needed += len(waveform_i)
                if not waveform_q in self.unique_waveforms:
                    memory_needed += len(waveform_q)

                if self.available_memory >= memory_needed:
                    if not waveform_i in self.unique_waveforms:
                        self.unique_waveforms.append(waveform_i)
                    if not waveform_q in self.unique_waveforms:
                        self.unique_waveforms.append(waveform_q)
                    self.available_memory -= memory_needed
                else:
                    raise WaveformsBuffer.NotEnoughMemory
        else:
            pulse.idx_range = self.bake_pulse_waveforms(pulse_copy, values, hardware_mod_en)

    def bake_pulse_waveforms(
        self, pulse: Pulse, values: list(), hardware_mod_en: bool
    ):  # bake_pulse_waveforms(self, pulse: Pulse, values: list(int), hardware_mod_en: bool):
        """Generates and stores a set of i and q waveforms required for a pulse duration sweep.

        These waveforms are generated and stored in a predefined order so that they can later be retrieved within the
        sweeper q1asm code. It bakes pulses from as short as 1ns, padding them at the end with 0s if required so that
        their length is a multiple of 4ns. It also supports the modulation of the pulse both in hardware (default)
        or software.
        With no other pulses stored in the sequencer memory, it supports values up to range(1, 126) for regular pulses and
        range(1, 180) for flux pulses.

        Args:
            pulse (:class:`qibolab.pulses.Pulse`): The pulse to be swept.
            values (list(int)): The list of values to sweep the pulse duration with.
            hardware_mod_en (bool): If set to True the pulses are assumed to be modulated in hardware and their
                envelope waveforms are uploaded; if False, software modulated waveforms are uploaded.

        Returns:
            idx_range (numpy.ndarray): An array with the indices of the set of pulses. For each pulse duration in
                `values` the i component is saved in the next avalable index, followed by the q component. For flux
                pulses, since both i and q components are equal, they are only saved once.

        Raises:
            NotEnoughMemory: If the memory needed to store the waveforms in more than the memory avalible.
        """
        # In order to generate waveforms for each duration value, the pulse will need to be modified.
        # To avoid any conflicts, make a copy of the pulse first.
        pulse_copy = pulse.copy()

        # there may be other waveforms stored already, set first index as the next available
        first_idx = len(self.unique_waveforms)

        if pulse.type == PulseType.FLUX:
            # for flux pulses, store i waveforms
            idx_range = np.arange(first_idx, first_idx + len(values), 1)

            for duration in values:
                pulse_copy.duration = duration
                if hardware_mod_en:
                    waveform = pulse_copy.envelope_waveform_i
                else:
                    waveform = pulse_copy.modulated_waveform_i

                padded_duration = int(np.ceil(duration / 4)) * 4
                memory_needed = padded_duration
                padding = np.zeros(padded_duration - duration)
                waveform.data = np.append(waveform.data, padding)

                if self.available_memory >= memory_needed:
                    self.unique_waveforms.append(waveform)
                    self.available_memory -= memory_needed
                else:
                    raise WaveformsBuffer.NotEnoughMemoryForBaking
        else:
            # for any other pulse type, store both i and q waveforms
            idx_range = np.arange(first_idx, first_idx + len(values) * 2, 2)

            for duration in values:
                pulse_copy.duration = duration
                if hardware_mod_en:
                    waveform_i, waveform_q = pulse_copy.envelope_waveforms
                else:
                    waveform_i, waveform_q = pulse_copy.modulated_waveforms

                padded_duration = int(np.ceil(duration / 4)) * 4
                memory_needed = padded_duration * 2
                padding = np.zeros(padded_duration - duration)
                waveform_i.data = np.append(waveform_i.data, padding)
                waveform_q.data = np.append(waveform_q.data, padding)

                if self.available_memory >= memory_needed:
                    self.unique_waveforms.append(waveform_i)
                    self.unique_waveforms.append(waveform_q)
                    self.available_memory -= memory_needed
                else:
                    raise WaveformsBuffer.NotEnoughMemoryForBaking

        return idx_range


class Sequencer:
    """A class to extend the functionality of qblox_instruments Sequencer.

    A sequencer is a hardware component synthesised in the instrument FPGA, responsible for fetching waveforms from
    memory, pre-processing them, sending them to the DACs, and processing the acquisitions from the ADCs (QRM modules).
    https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/documentation/sequencer.html

    This class extends the sequencer functionality by holding additional data required when
    processing a pulse sequence:

    - the sequencer number,
    - the sequence of pulses to be played,
    - a buffer of unique waveforms, and
    - the four components of the sequence file:

        - waveforms dictionary
        - acquisition dictionary
        - weights dictionary
        - program

    Attributes:
        device (QbloxSequencer): A reference to the underlying `qblox_instruments.qcodes_drivers.sequencer.Sequencer`
            object. It can be used to access other features not directly exposed by this wrapper.
            https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/documentation/sequencer.html
        number (int): An integer between 0 and 5 that identifies the number of the sequencer.
        pulses (PulseSequence): The sequence of pulses to be played by the sequencer.
        waveforms_buffer (WaveformsBuffer): A buffer of unique waveforms to be played by the sequencer.
        waveforms (dict): A dictionary containing the waveforms to be played by the sequencer in qblox format.
        acquisitions (dict): A dictionary containing the list of acquisitions to be made by the sequencer in qblox
            format.
        weights (dict): A dictionary containing the list of weights to be used by the sequencer when demodulating
            and integrating the response, in qblox format.
        program (str): The pseudo assembly (q1asm) program to be executed by the sequencer.
            https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/documentation/sequencer.html#instructions

        qubit (str): The id of the qubit associated with the sequencer, if there is only one.
    """

    def __init__(self, number: int):
        """Initialises the sequencer.

        All class attributes are defined and initialised.
        """

        self.device: QbloxSequencer = None
        self.number: int = number
        self.pulses: PulseSequence = PulseSequence()
        self.waveforms_buffer: WaveformsBuffer = WaveformsBuffer()
        self.waveforms: dict = {}
        self.acquisitions: dict = {}
        self.weights: dict = {}
        self.program: Program = Program()
        self.qubit = None  # self.qubit: int | str = None



"""Helper methods for loading and saving to runcards.

The format of runcards in the ``qiboteam/qibolab_platforms_qrc``
repository is assumed here. See :ref:`Using runcards <using_runcards>`
example for more details.
"""
from dataclasses import asdict
from pathlib import Path
from typing import Tuple

import yaml

from qibolab.native import SingleQubitNatives, TwoQubitNatives
from qibolab.platform import Platform, QubitMap, QubitPairMap, Settings
from qibolab.qubits import Qubit, QubitPair


def load_runcard(path: Path) -> dict:
    """Load runcard YAML to a dictionary."""
    return yaml.safe_load(path.read_text())


def load_settings(runcard: dict) -> Settings:
    """Load platform settings section from the runcard."""
    return Settings(**runcard["settings"])


def load_qubits(runcard: dict) -> Tuple[QubitMap, QubitPairMap]:
    """Load qubits and pairs from the runcard.

    Uses the native gate and characterization sections of the runcard
    to parse the :class:`qibolab.qubits.Qubit` and :class:`qibolab.qubits.QubitPair`
    objects.
    """
    qubits = {q: Qubit(q, **char) for q, char in runcard["characterization"]["single_qubit"].items()}

    pairs = {}
    for pair in runcard["topology"]:
        pair = tuple(sorted(pair))
        pairs[pair] = QubitPair(qubits[pair[0]], qubits[pair[1]])

    # register single qubit native gates to ``Qubit`` objects
    native_gates = runcard.get("native_gates", {})
    for q, gates in native_gates.get("single_qubit", {}).items():
        qubits[q].native_gates = SingleQubitNatives.from_dict(qubits[q], gates)
    # register two-qubit native gates to ``QubitPair`` objects
    for pair, gatedict in native_gates.get("two_qubit", {}).items():
        pair = tuple(sorted(int(q) if q.isdigit() else q for q in pair.split("-")))
        pairs[pair].native_gates = TwoQubitNatives.from_dict(qubits, gatedict)

    return qubits, pairs


def dump_qubits(qubits: QubitMap, pairs: QubitPairMap) -> dict:
    """Dump qubit and pair objects to a dictionary following the runcard format."""
    native_gates = {
        "single_qubit": {q: qubit.native_gates.raw for q, qubit in qubits.items()},
        "two_qubit": {},
    }
    # add two-qubit native gates
    for p, pair in pairs.items():
        natives = pair.native_gates.raw
        if len(natives) > 0:
            native_gates["two_qubit"][f"{p[0]}-{p[1]}"] = natives
    # add qubit characterization section
    return {
        "native_gates": native_gates,
        "characterization": {"single_qubit": {q: qubit.characterization for q, qubit in qubits.items()}},
    }


def dump_runcard(platform: Platform, path: Path):
    """Serializes the platform and saves it as a yaml runcard file.

    The file saved follows the format explained in :ref:`Using runcards <using_runcards>`.

    Args:
        platform (qibolab.platform.Platform): The platform to be serialized.
        path (pathlib.Path): Path that the yaml file will be saved.
    """
    settings = {
        "nqubits": platform.nqubits,
        "qubits": list(platform.qubits),
        "settings": asdict(platform.settings),
        "topology": [list(pair) for pair in platform.pairs],
    }
    settings.update(dump_qubits(platform.qubits, platform.pairs))
    path.write_text(yaml.dump(settings, sort_keys=False, indent=4, default_flow_style=None))



from qibolab.transpilers.fusion import Fusion, Rearrange
from qibolab.transpilers.gate_decompositions import NativeGates
from qibolab.transpilers.pipeline import Pipeline
from qibolab.transpilers.placer import Random, ReverseTraversal, Subgraph, Trivial
from qibolab.transpilers.routing import ShortestPaths
from qibolab.transpilers.star_connectivity import StarConnectivity



"""Convert helper functions for rfsoc driver."""
from dataclasses import asdict
from functools import singledispatch

import numpy as np
import qibosoq.components.base as rfsoc
import qibosoq.components.pulses as rfsoc_pulses

from qibolab.platform import Qubit
from qibolab.pulses import Pulse, PulseSequence, PulseShape
from qibolab.sweeper import BIAS, DURATION, START, Parameter, Sweeper

HZ_TO_MHZ = 1e-6
NS_TO_US = 1e-3


def replace_pulse_shape(rfsoc_pulse: rfsoc_pulses.Pulse, shape: PulseShape) -> rfsoc_pulses.Pulse:
    """Set pulse shape parameters in rfsoc_pulses pulse object."""
    if shape.name not in {"Gaussian", "Drag", "Rectangular", "Exponential"}:
        new_pulse = rfsoc_pulses.Arbitrary(
            **asdict(rfsoc_pulse), i_values=shape.envelope_waveform_i, q_values=shape.envelope_waveform_q
        )
        return new_pulse
    new_pulse_cls = getattr(rfsoc_pulses, shape.name)
    if shape.name == "Rectangular":
        return new_pulse_cls(**asdict(rfsoc_pulse))
    if shape.name == "Gaussian":
        return new_pulse_cls(**asdict(rfsoc_pulse), rel_sigma=shape.rel_sigma)
    if shape.name == "Drag":
        return new_pulse_cls(**asdict(rfsoc_pulse), rel_sigma=shape.rel_sigma, beta=shape.beta)
    if shape.name == "Exponential":
        return new_pulse_cls(**asdict(rfsoc_pulse), tau=shape.tau, upsilon=shape.upsilon, weight=shape.g)


def pulse_lo_frequency(pulse: Pulse, qubits: dict[int, Qubit]) -> int:
    """Return local_oscillator frequency (HZ) of a pulse."""
    pulse_type = pulse.type.name.lower()
    try:
        lo_frequency = getattr(qubits[pulse.qubit], pulse_type).local_oscillator._frequency
    except AttributeError:
        lo_frequency = 0
    return lo_frequency


def convert_units_sweeper(sweeper: rfsoc.Sweeper, sequence: PulseSequence, qubits: dict[int, Qubit]):
    """Convert units for `qibosoq.abstract.Sweeper` considering also LOs."""
    for idx, jdx in enumerate(sweeper.indexes):
        parameter = sweeper.parameters[idx]
        if parameter is rfsoc.Parameter.FREQUENCY:
            pulse = sequence[jdx]
            lo_frequency = pulse_lo_frequency(pulse, qubits)
            sweeper.starts[idx] = (sweeper.starts[idx] - lo_frequency) * HZ_TO_MHZ
            sweeper.stops[idx] = (sweeper.stops[idx] - lo_frequency) * HZ_TO_MHZ
        elif parameter is rfsoc.Parameter.DELAY:
            sweeper.starts[idx] *= NS_TO_US
            sweeper.stops[idx] *= NS_TO_US
        elif parameter is rfsoc.Parameter.RELATIVE_PHASE:
            sweeper.starts[idx] = np.degrees(sweeper.starts[idx])
            sweeper.stops[idx] = np.degrees(sweeper.stops[idx])


@singledispatch
def convert(*args):
    """Convert from qibolab obj to qibosoq obj, overloaded."""
    raise ValueError(f"Convert function received bad parameters ({type(args[0])}).")


@convert.register
def _(qubit: Qubit) -> rfsoc.Qubit:
    """Convert `qibolab.platforms.abstract.Qubit` to `qibosoq.abstract.Qubit`."""
    if qubit.flux:
        return rfsoc.Qubit(qubit.flux.offset, qubit.flux.port.name)
    return rfsoc.Qubit(0.0, None)


@convert.register
def _(sequence: PulseSequence, qubits: dict[int, Qubit]) -> list[rfsoc_pulses.Pulse]:
    """Convert PulseSequence to list of rfosc pulses with relative time."""
    last_pulse_start = 0
    list_sequence = []
    for pulse in sorted(sequence.pulses, key=lambda item: item.start):
        start_delay = (pulse.start - last_pulse_start) * NS_TO_US
        pulse_dict = asdict(convert(pulse, qubits, start_delay))
        list_sequence.append(pulse_dict)

        last_pulse_start = pulse.start
    return list_sequence


@convert.register
def _(pulse: Pulse, qubits: dict[int, Qubit], start_delay: float) -> rfsoc_pulses.Pulse:
    """Convert `qibolab.pulses.pulse` to `qibosoq.abstract.Pulse`."""
    pulse_type = pulse.type.name.lower()
    dac = getattr(qubits[pulse.qubit], pulse_type).port.name
    adc = qubits[pulse.qubit].feedback.port.name if pulse_type == "readout" else None
    lo_frequency = pulse_lo_frequency(pulse, qubits)

    rfsoc_pulse = rfsoc_pulses.Pulse(
        frequency=(pulse.frequency - lo_frequency) * HZ_TO_MHZ,
        amplitude=pulse.amplitude,
        relative_phase=np.degrees(pulse.relative_phase),
        start_delay=start_delay,
        duration=pulse.duration * NS_TO_US,
        dac=dac,
        adc=adc,
        name=pulse.serial,
        type=pulse_type,
    )
    return replace_pulse_shape(rfsoc_pulse, pulse.shape)


@convert.register
def _(par: Parameter) -> rfsoc.Parameter:
    """Convert a qibolab sweeper.Parameter into a qibosoq.Parameter."""
    return getattr(rfsoc.Parameter, par.name.upper())


@convert.register
def _(sweeper: Sweeper, sequence: PulseSequence, qubits: dict[int, Qubit]) -> rfsoc.Sweeper:
    """Convert `qibolab.sweeper.Sweeper` to `qibosoq.abstract.Sweeper`.

    Note that any unit conversion is not done in this function (to avoid to do it multiple times).
    Conversion will be done in `convert_units_sweeper`.
    """
    parameters = []
    starts = []
    stops = []
    indexes = []

    if sweeper.parameter is BIAS:
        for qubit in sweeper.qubits:
            parameters.append(rfsoc.Parameter.BIAS)
            indexes.append(list(qubits.values()).index(qubit))

            base_value = qubit.flux.offset
            values = sweeper.get_values(base_value)
            starts.append(values[0])
            stops.append(values[-1])

        if max(np.abs(starts)) > 1 or max(np.abs(stops)) > 1:
            raise ValueError("Sweeper amplitude is set to reach values higher than 1")
    else:
        for pulse in sweeper.pulses:
            idx_sweep = sequence.index(pulse)
            indexes.append(idx_sweep)
            base_value = getattr(pulse, sweeper.parameter.name)
            if idx_sweep != 0 and sweeper.parameter is START:
                # do the conversion from start to delay
                base_value = base_value - sequence[idx_sweep - 1].start
            values = sweeper.get_values(base_value)
            starts.append(values[0])
            stops.append(values[-1])

            if sweeper.parameter is START:
                parameters.append(rfsoc.Parameter.DELAY)
            elif sweeper.parameter is DURATION:
                parameters.append(rfsoc.Parameter.DURATION)
                delta_start = values[0] - base_value
                delta_stop = values[-1] - base_value

                if len(sequence) > idx_sweep + 1:
                    # if duration-swept pulse is not last
                    indexes.append(idx_sweep + 1)
                    t_start = sequence[idx_sweep + 1].start - sequence[idx_sweep].start
                    parameters.append(rfsoc.Parameter.DELAY)
                    starts.append(t_start + delta_start)
                    stops.append(t_start + delta_stop)
            else:
                parameters.append(convert(sweeper.parameter))

    return rfsoc.Sweeper(
        parameters=parameters,
        indexes=indexes,
        starts=starts,
        stops=stops,
        expts=len(sweeper.values),
    )



from dataclasses import dataclass

from qibo import gates

from qibolab.transpilers.abstract import Transpiler


@dataclass
class Fusion(Transpiler):
    """Apply gate fusion up to the given ``max_qubits``."""

    max_qubits: int = 1

    def is_satisfied(self, circuit):
        return True

    def __call__(self, circuit):
        return circuit.fuse(max_qubits=self.max_qubits), list(range(circuit.nqubits))


@dataclass
class Rearrange(Fusion):
    """Rearranges gates using qibo's fusion algorithm.

    May reduce number of SWAPs when fixing for connectivity
    but this has not been tested.
    """

    def __call__(self, circuit):
        fcircuit = circuit.fuse(max_qubits=self.max_qubits)
        new = circuit.__class__(circuit.nqubits)
        for fgate in fcircuit.queue:
            if isinstance(fgate, gates.FusedGate):
                new.add(fgate.gates)
            else:
                new.add(fgate)
        return new, list(range(circuit.nqubits))



""" Qblox Cluster QCM driver."""

import json

from qblox_instruments.qcodes_drivers.cluster import Cluster as QbloxCluster
from qblox_instruments.qcodes_drivers.qcm_qrm import QcmQrm as QbloxQrmQcm
from qibo.config import log

from qibolab.instruments.abstract import Instrument
from qibolab.instruments.qblox.port import (
    ClusterBB_OutputPort,
    ClusterBB_OutputPort_Settings,
)
from qibolab.instruments.qblox.q1asm import (
    Block,
    Register,
    convert_phase,
    loop_block,
    wait_block,
)
from qibolab.instruments.qblox.sequencer import Sequencer, WaveformsBuffer
from qibolab.instruments.qblox.sweeper import QbloxSweeper, QbloxSweeperType
from qibolab.pulses import Pulse, PulseSequence, PulseType
from qibolab.sweeper import Parameter, Sweeper, SweeperType


class ClusterQCM_BB_Settings:
    def __init__(self, ports=None):
        if not ports:
            ports: dict = {
                "o1": ClusterBB_OutputPort_Settings(),
                "o2": ClusterBB_OutputPort_Settings(),
                "o3": ClusterBB_OutputPort_Settings(),
                "o4": ClusterBB_OutputPort_Settings(),
            }
        self.ports = ports


class ClusterQCM_BB(Instrument):
    """Qblox Cluster Qubit Control Module Baseband driver.

    Qubit Control Module (QCM) is an arbitratry wave generator with two DACs connected to
    four output ports. It can sinthesise either four independent real signals or two
    complex signals, using ports 0 and 2 to output the i(in-phase) component and
    ports 1 and 3 the q(quadrature) component. The sampling rate of its DAC is 1 GSPS.
    https://www.qblox.com/cluster

    The class aims to simplify the configuration of the instrument, exposing only
    those parameters most frequencly used and hiding other more complex components.

    A reference to the underlying `qblox_instruments.qcodes_drivers.qcm_qrm.QRM_QCM`
    object is provided via the attribute `device`, allowing the advanced user to gain
    access to the features that are not exposed directly by the class.

    In order to accelerate the execution, the instrument settings are cached, so that
    the communication with the instrument only happens when the parameters change.
    This caching is done with the method `_set_device_parameter(target, *parameters, value)`.

    .. code-block:: text

            ports:
                o1:
                    channel                      : L4-1
                    gain                         : 0.2 # -1.0<=v<=1.0
                    offset                       : 0   # -2.5<=v<=2.5
                o2:
                    channel                      : L4-2
                    gain                         : 0.2 # -1.0<=v<=1.0
                    offset                       : 0   # -2.5<=v<=2.5
                o3:
                    channel                      : L4-3
                    gain                         : 0.2 # -1.0<=v<=1.0
                    offset                       : 0   # -2.5<=v<=2.5
                o4:
                    channel                      : L4-4
                    gain                         : 0.2 # -1.0<=v<=1.0
                    offset                       : 0   # -2.5<=v<=2.5

    Attributes:
        name (str): A unique name given to the instrument.
        address (str): IP_address:module_number (the IP address of the cluster and
            the module number)
        device (QbloxQrmQcm): A reference to the underlying
            `qblox_instruments.qcodes_drivers.qcm_qrm.QcmQrm` object. It can be used to access other
            features not directly exposed by this wrapper.
            https://qblox-qblox-instruments.readthedocs-hosted.com/en/master/api_reference/qcm_qrm.html
        ports = A dictionary giving access to the output ports objects.

            - ports['o1']
            - ports['o2']
            - ports['o3']
            - ports['o4']

            - ports['oX'].channel (int | str): the id of the refrigerator channel the port is connected to.
            - ports['oX'].gain (float): (mapped to qrm.sequencers[0].gain_awg_path0 and qrm.sequencers[0].gain_awg_path1)
                Sets the gain on both paths of the output port.
            - ports['oX'].offset (float): (mapped to qrm.outX_offset)
                Sets the offset on the output port.
            - ports['oX'].hardware_mod_en (bool): (mapped to qrm.sequencers[0].mod_en_awg) Enables pulse
                modulation in hardware. When set to False, pulse modulation is done at the host computer
                and a modulated pulse waveform should be uploaded to the instrument. When set to True,
                the envelope of the pulse should be uploaded to the instrument and it modulates it in
                real time by its FPGA using the sequencer nco (numerically controlled oscillator).
            - ports['oX'].nco_freq (int): (mapped to qrm.sequencers[0].nco_freq)        # TODO mapped, but not configurable from the runcard
            - ports['oX'].nco_phase_offs = (mapped to qrm.sequencers[0].nco_phase_offs) # TODO mapped, but not configurable from the runcard

                - Sequencer 0 is always the first sequencer used to synthesise pulses on port o1.
                - Sequencer 1 is always the first sequencer used to synthesise pulses on port o2.
                - Sequencer 2 is always the first sequencer used to synthesise pulses on port o3.
                - Sequencer 3 is always the first sequencer used to synthesise pulses on port o4.
                - Sequencer 4 to 6 are used as needed to sinthesise simultaneous pulses on the same channel
                  or when the memory of the default sequencers rans out.
    """

    DEFAULT_SEQUENCERS = {"o1": 0, "o2": 1, "o3": 2, "o4": 3}
    SAMPLING_RATE: int = 1e9  # 1 GSPS
    FREQUENCY_LIMIT = 500e6

    property_wrapper = lambda parent, *parameter: property(
        lambda self: parent.device.get(parameter[0]),
        lambda self, x: parent._set_device_parameter(parent.device, *parameter, value=x),
    )
    sequencer_property_wrapper = lambda parent, sequencer, *parameter: property(
        lambda self: parent.device.sequencers[sequencer].get(parameter[0]),
        lambda self, x: parent._set_device_parameter(parent.device.sequencers[sequencer], *parameter, value=x),
    )

    def __init__(self, name: str, address: str, settings: dict):
        """Initialises the instance.

        All class attributes are defined and initialised.
        """
        super().__init__(name, address)
        self.settings: ClusterQCM_BB_Settings = settings
        self.device: QbloxQrmQcm = None
        self.ports: dict = {}
        for n in range(4):
            port = "o" + str(n + 1)
            self.ports[port] = ClusterBB_OutputPort(
                module=self, sequencer_number=self.DEFAULT_SEQUENCERS[port], number=n + 1
            )
        self.channels: list = []

        self._debug_folder: str = ""
        self._cluster: QbloxCluster = None
        self._output_ports_keys = ["o1", "o2", "o3", "o4"]
        self._sequencers: dict[Sequencer] = {"o1": [], "o2": [], "o3": [], "o4": []}
        self._port_channel_map: dict = {}
        self._channel_port_map: dict = {}
        self._device_parameters = {}
        self._device_num_output_ports = 2
        self._device_num_sequencers: int
        self._free_sequencers_numbers: list[int] = []
        self._used_sequencers_numbers: list[int] = []
        self._unused_sequencers_numbers: list[int] = []

    def connect(self, cluster: QbloxCluster):
        """Connects to the instrument using the instrument settings in the runcard.

        Once connected, it creates port classes with properties mapped to various instrument
        parameters, and initialises the the underlying device parameters.
        """
        if not self.is_connected:
            if cluster:
                # save a reference to the underlying object
                self.device = cluster.modules[int(self.address.split(":")[1]) - 1]
                # TODO: test connection with the module before continuing

                # save reference to cluster
                self._cluster = cluster
                self.is_connected = True

                # once connected, initialise the parameters of the device to the default values
                self._set_device_parameter(self.device, "out0_offset", value=0)  # Default after reboot = 0
                self._set_device_parameter(self.device, "out1_offset", value=0)  # Default after reboot = 0
                self._set_device_parameter(self.device, "out2_offset", value=0)  # Default after reboot = 0
                self._set_device_parameter(self.device, "out3_offset", value=0)  # Default after reboot = 0

                # initialise the parameters of the default sequencers to the default values,
                # the rest of the sequencers are not configured here, but will be configured
                # with the same parameters as the default in process_pulse_sequence()
                for target in [
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o1"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o2"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o3"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o4"]],
                ]:
                    self._set_device_parameter(
                        target, "cont_mode_en_awg_path0", "cont_mode_en_awg_path1", value=False
                    )  # Default after reboot = False
                    self._set_device_parameter(
                        target, "cont_mode_waveform_idx_awg_path0", "cont_mode_waveform_idx_awg_path1", value=0
                    )  # Default after reboot = 0
                    self._set_device_parameter(target, "marker_ovr_en", value=True)  # Default after reboot = False
                    self._set_device_parameter(target, "marker_ovr_value", value=15)  # Default after reboot = 0
                    self._set_device_parameter(target, "mixer_corr_gain_ratio", value=1)
                    self._set_device_parameter(target, "mixer_corr_phase_offset_degree", value=0)
                    self._set_device_parameter(target, "offset_awg_path0", value=0)
                    self._set_device_parameter(target, "offset_awg_path1", value=0)
                    self._set_device_parameter(target, "sync_en", value=False)  # Default after reboot = False
                    self._set_device_parameter(target, "upsample_rate_awg_path0", "upsample_rate_awg_path1", value=0)

                for target in [
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o1"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o2"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o3"]],
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o4"]],
                ]:
                    self._set_device_parameter(
                        target, "channel_map_path0_out0_en", value=False
                    )  # Default after reboot = True
                    self._set_device_parameter(
                        target, "channel_map_path1_out1_en", value=False
                    )  # Default after reboot = True
                    self._set_device_parameter(
                        target, "channel_map_path0_out2_en", value=False
                    )  # Default after reboot = True
                    self._set_device_parameter(
                        target, "channel_map_path1_out3_en", value=False
                    )  # Default after reboot = True

                self._set_device_parameter(
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o1"]],
                    "channel_map_path0_out0_en",
                    value=True,
                )
                self._set_device_parameter(
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o2"]],
                    "channel_map_path1_out1_en",
                    value=True,
                )
                self._set_device_parameter(
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o3"]],
                    "channel_map_path0_out2_en",
                    value=True,
                )
                self._set_device_parameter(
                    self.device.sequencers[self.DEFAULT_SEQUENCERS["o4"]],
                    "channel_map_path1_out3_en",
                    value=True,
                )

                # on initialisation, disconnect all other sequencers from the ports
                self._device_num_sequencers = len(self.device.sequencers)
                for sequencer in range(4, self._device_num_sequencers):
                    target = self.device.sequencers[sequencer]
                    self._set_device_parameter(target, "channel_map_path0_out0_en", value=False)
                    self._set_device_parameter(target, "channel_map_path1_out1_en", value=False)
                    self._set_device_parameter(target, "channel_map_path0_out2_en", value=False)
                    self._set_device_parameter(target, "channel_map_path1_out3_en", value=False)

    def _set_device_parameter(self, target, *parameters, value):
        """Sets a parameter of the instrument, if it changed from the last stored in the cache.

        Args:
            target = an instance of qblox_instruments.qcodes_drivers.qcm_qrm.QcmQrm or
                                    qblox_instruments.qcodes_drivers.sequencer.Sequencer
            *parameters (list): A list of parameters to be cached and set.
            value = The value to set the paramters.
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        if self.is_connected:
            key = target.name + "." + parameters[0]
            if not key in self._device_parameters:
                for parameter in parameters:
                    if not hasattr(target, parameter):
                        raise Exception(f"The instrument {self.name} does not have parameters {parameter}")
                    target.set(parameter, value)
                self._device_parameters[key] = value
            elif self._device_parameters[key] != value:
                for parameter in parameters:
                    target.set(parameter, value)
                self._device_parameters[key] = value
        else:
            raise Exception("There is no connection to the instrument {self.name}")

    def _erase_device_parameters_cache(self):
        """Erases the cache of instrument parameters."""
        self._device_parameters = {}

    def setup(self):
        """Configures the instrument with the settings of the runcard.

        A connection to the instrument needs to be established beforehand.
        Args:
            **kwargs: dict = A dictionary of settings loaded from the runcard:

                - oX: ['o1', 'o2', 'o3', 'o4']
                - kwargs['ports']['oX']['channel'] (int | str): the id of the refrigerator channel the port is connected to.
                - kwargs['ports'][oX]['gain'] (float): [0.0 - 1.0 unitless] gain applied prior to up-conversion. Qblox recommends to keep
                  `pulse_amplitude * gain` below 0.3 to ensure the mixers are working in their linear regime, if necessary, lowering the attenuation
                  applied at the output.
                - kwargs['ports'][oX]['offset'] (float): [-2.5 - 2.5 V] offset in volts applied to the output port.
                - kwargs['ports'][oX]['hardware_mod_en'] (bool): enables Hardware Modulation. In this mode, pulses are modulated to the intermediate frequency
                  using the numerically controlled oscillator within the fpga. It only requires the upload of the pulse envelope waveform.

        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        settings: ClusterQCM_BB_Settings = self.settings
        if self.is_connected:
            # Load settings
            for port in ["o1", "o2", "o3", "o4"]:
                if port in settings.ports:
                    port_settings: ClusterBB_OutputPort_Settings = settings.ports[port]
                    self.ports[port].channel = port_settings.channel
                    self._port_channel_map[port] = self.ports[port].channel
                    self.ports[port].gain = port_settings.gain
                    self.ports[port].offset = port_settings.offset
                    self.ports[port].hardware_mod_en = port_settings.hardware_mod_en
                    self.ports[port].qubit = port_settings.qubit
                    self.ports[port].nco_freq = 0
                    self.ports[port].nco_phase_offs = 0
                else:
                    if port in self.ports:
                        self.ports[port].channel = None
                        self.ports[port].gain = 0
                        self.ports[port].offset = 0
                        self.ports[port].hardware_mod_en = False
                        self.ports[port].qubit = None
                        self.ports[port].nco_freq = 0
                        self.ports[port].nco_phase_offs = 0
                        self.ports.pop(port)
                        self._output_ports_keys.remove(port)
                        self._sequencers.pop(port)

            self._channel_port_map = {v: k for k, v in self._port_channel_map.items()}
            self.channels = list(self._channel_port_map.keys())
        else:
            raise Exception("The instrument cannot be set up, there is no connection")

    def _get_next_sequencer(self, port, frequency, qubits: dict, qubit: None):
        """Retrieves and configures the next avaliable sequencer.

        The parameters of the new sequencer are copied from those of the default sequencer, except for the
        intermediate frequency and classification parameters.
        Args:
            port (str):
            frequency ():
            qubit ():
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """

        # select a new sequencer and configure it as required
        next_sequencer_number = self._free_sequencers_numbers.pop(0)
        if next_sequencer_number != self.DEFAULT_SEQUENCERS[port]:
            for parameter in self.device.sequencers[self.DEFAULT_SEQUENCERS[port]].parameters:
                # exclude read-only parameter `sequence`
                if not parameter in ["sequence"]:
                    value = self.device.sequencers[self.DEFAULT_SEQUENCERS[port]].get(param_name=parameter)
                    if value:
                        target = self.device.sequencers[next_sequencer_number]
                        self._set_device_parameter(target, parameter, value=value)

        # if hardware modulation is enabled configure nco_frequency
        if self.ports[port].hardware_mod_en:
            self._set_device_parameter(
                self.device.sequencers[next_sequencer_number],
                "nco_freq",
                value=frequency,  # Assumes all pulses in non_overlapping_pulses set
                # have the same frequency. Non-overlapping pulses of different frequencies on the same
                # qubit channel with hardware_demod_en would lead to wrong results.
                # TODO: Throw error in that event or implement for non_overlapping_same_frequency_pulses
                #       Even better, set the frequency before each pulse is played (would work with hardware modulation only)
            )

            # set sequencer port offset to the runcard value
            # self._set_device_parameter(
            #     self.device.sequencers[next_sequencer_number],
            #     "out0_offset", #How to select device parameter offset??
            #     value=qubits[qubit].sweetspot,
            # )

            self.ports[port].offset = qubits[qubit].sweetspot

        # create sequencer wrapper
        sequencer = Sequencer(next_sequencer_number)
        sequencer.qubit = qubit
        return sequencer

    def get_if(self, pulse):
        """Returns the intermediate frequency needed to synthesise a pulse based on the port lo frequency."""

        _rf = pulse.frequency
        _lo = 0  # QCMs do not have local oscillator
        _if = _rf - _lo
        if abs(_if) > self.FREQUENCY_LIMIT:
            raise RuntimeError(
                f"""
            Pulse frequency {_rf:_} cannot be synthesised, it exceeds the maximum frequency of {self.FREQUENCY_LIMIT:_}"""
            )
        return _if

    def process_pulse_sequence(
        self,
        qubits: dict,
        instrument_pulses: PulseSequence,
        navgs: int,
        nshots: int,
        repetition_duration: int,
        sweepers=None,
    ):
        """Processes a list of pulses, generating the waveforms and sequence program required by
        the instrument to synthesise them.

        The output of the process is a list of sequencers used for each port, configured with the information
        required to play the sequence.
        The following features are supported:

        - overlapping pulses
        - hardware modulation
        - software modulation, with support for arbitrary pulses
        - real-time sweepers of

            - pulse frequency (requires hardware modulation)
            - pulse relative phase (requires hardware modulation)
            - pulse amplitude
            - pulse start
            - pulse duration
            - port gain
            - port offset

        - sequencer memory optimisation (waveforms cache)
        - extended waveform memory with the use of multiple sequencers
        - pulses of up to 8192 pairs of i, q samples
        - intrument parameters cache

        Args:
            instrument_pulses (PulseSequence): A collection of Pulse objects to be played by the instrument.
            navgs (int): The number of times the sequence of pulses should be executed averaging the results.
            nshots (int): The number of times the sequence of pulses should be executed without averaging.
            repetition_duration (int): The total duration of the pulse sequence execution plus the reset/relaxation time.
            sweepers (list(Sweeper)): A list of Sweeper objects to be implemented.
        """
        if sweepers is None:
            sweepers = []
        sequencer: Sequencer
        sweeper: Sweeper

        self._free_sequencers_numbers = list(range(len(self.ports), 6))

        # process the pulses for every port
        for port in self.ports:
            # split the collection of instruments pulses by ports
            port_pulses: PulseSequence = instrument_pulses.get_channel_pulses(self._port_channel_map[port])

            # initialise the list of sequencers required by the port
            self._sequencers[port] = []

            # initialise the list of free sequencer numbers to include the default for each port {'o1': 0, 'o2': 1, 'o3': 2, 'o4': 3}
            self._free_sequencers_numbers = [self.DEFAULT_SEQUENCERS[port]] + self._free_sequencers_numbers

            if not port_pulses.is_empty:
                # split the collection of port pulses in non overlapping pulses
                non_overlapping_pulses: PulseSequence
                for non_overlapping_pulses in port_pulses.separate_overlapping_pulses():
                    # TODO: for non_overlapping_same_frequency_pulses in non_overlapping_pulses.separate_different_frequency_pulses():

                    # each set of not overlapping pulses will be played by a separate sequencer
                    # check sequencer availability
                    if len(self._free_sequencers_numbers) == 0:
                        raise Exception(
                            f"The number of sequencers requried to play the sequence exceeds the number available {self._device_num_sequencers}."
                        )
                    # get next sequencer
                    sequencer = self._get_next_sequencer(
                        port=port,
                        frequency=self.get_if(non_overlapping_pulses[0]),
                        qubits=qubits,
                        qubit=non_overlapping_pulses[0].qubit,
                    )
                    # add the sequencer to the list of sequencers required by the port
                    self._sequencers[port].append(sequencer)

                    # make a temporary copy of the pulses to be processed
                    pulses_to_be_processed = non_overlapping_pulses.shallow_copy()
                    while not pulses_to_be_processed.is_empty:
                        pulse: Pulse = pulses_to_be_processed[0]
                        # attempt to save the waveforms to the sequencer waveforms buffer
                        try:
                            sequencer.waveforms_buffer.add_waveforms(pulse, self.ports[port].hardware_mod_en, sweepers)
                            sequencer.pulses.add(pulse)
                            pulses_to_be_processed.remove(pulse)

                        # if there is not enough memory in the current sequencer, use another one
                        except WaveformsBuffer.NotEnoughMemory:
                            if len(pulse.waveform_i) + len(pulse.waveform_q) > WaveformsBuffer.SIZE:
                                raise NotImplementedError(
                                    f"Pulses with waveforms longer than the memory of a sequencer ({WaveformsBuffer.SIZE // 2}) are not supported."
                                )
                            if len(self._free_sequencers_numbers) == 0:
                                raise Exception(
                                    f"The number of sequencers requried to play the sequence exceeds the number available {self._device_num_sequencers}."
                                )
                            # get next sequencer
                            sequencer = self._get_next_sequencer(
                                port=port,
                                frequency=self.get_if(non_overlapping_pulses[0]),
                                qubits=qubits,
                                qubit=non_overlapping_pulses[0].qubit,
                            )
                            # add the sequencer to the list of sequencers required by the port
                            self._sequencers[port].append(sequencer)
            else:
                sequencer = self._get_next_sequencer(
                    port=port, frequency=0, qubits=qubits, qubit=self.ports[port].qubit
                )
                # add the sequencer to the list of sequencers required by the port
                self._sequencers[port].append(sequencer)

        # update the lists of used and unused sequencers that will be needed later on
        self._used_sequencers_numbers = []
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                self._used_sequencers_numbers.append(sequencer.number)
        self._unused_sequencers_numbers = []
        for n in range(self._device_num_sequencers):
            if not n in self._used_sequencers_numbers:
                self._unused_sequencers_numbers.append(n)

        # generate and store the Waveforms dictionary, the Acquisitions dictionary, the Weights and the Program
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                pulses = sequencer.pulses
                program = sequencer.program

                ## pre-process sweepers ##
                # TODO: move qibolab sweepers preprocessing to qblox controller

                # attach a sweeper attribute to the pulse so that it is easily accesible by the code that generates
                # the pseudo-assembly program
                pulse = None
                for pulse in pulses:
                    pulse.sweeper = None

                pulse_sweeper_parameters = [
                    Parameter.frequency,
                    Parameter.amplitude,
                    Parameter.duration,
                    Parameter.relative_phase,
                    Parameter.start,
                ]

                for sweeper in sweepers:
                    if sweeper.parameter in pulse_sweeper_parameters:
                        # check if this sequencer takes an active role in the sweep
                        if sweeper.pulses and set(sequencer.pulses) & set(sweeper.pulses):
                            # plays an active role
                            reference_value = None
                            if sweeper.parameter == Parameter.frequency:
                                if sequencer.pulses:
                                    reference_value = self.get_if(
                                        sequencer.pulses[0]
                                    )  # uses the frequency of the first pulse (assuming all same freq)
                            if sweeper.parameter == Parameter.amplitude:
                                for pulse in pulses:
                                    if pulse in sweeper.pulses:
                                        reference_value = pulse.amplitude  # uses the amplitude of the first pulse
                            if sweeper.parameter == Parameter.duration and pulse in sweeper.pulses:
                                # for duration sweepers bake waveforms
                                sweeper.qs = QbloxSweeper(
                                    program=program, type=QbloxSweeperType.duration, rel_values=pulse.idx_range
                                )
                            else:
                                # create QbloxSweepers and attach them to qibolab sweeper
                                if sweeper.type == SweeperType.OFFSET and reference_value:
                                    sweeper.qs = QbloxSweeper.from_sweeper(
                                        program=program, sweeper=sweeper, add_to=reference_value
                                    )
                                elif sweeper.type == SweeperType.FACTOR and reference_value:
                                    sweeper.qs = QbloxSweeper.from_sweeper(
                                        program=program, sweeper=sweeper, multiply_to=reference_value
                                    )
                                else:
                                    sweeper.qs = QbloxSweeper.from_sweeper(program=program, sweeper=sweeper)

                            # finally attach QbloxSweepers to the pulses being swept
                            sweeper.qs.update_parameters = True
                            pulse.sweeper = sweeper.qs
                        else:
                            # does not play an active role
                            sweeper.qs = QbloxSweeper(
                                program=program,
                                type=QbloxSweeperType.number,
                                rel_values=range(len(sweeper.values)),
                                name=sweeper.parameter.name,
                            )

                    else:  # qubit_sweeper_parameters
                        if sweeper.qubits and sequencer.qubit in [_.name for _ in sweeper.qubits]:
                            # plays an active role
                            if sweeper.parameter == Parameter.bias:
                                reference_value = self.ports[port].offset
                                # create QbloxSweepers and attach them to qibolab sweeper
                                if sweeper.type == SweeperType.ABSOLUTE:
                                    sweeper.qs = QbloxSweeper.from_sweeper(
                                        program=program, sweeper=sweeper, add_to=-reference_value
                                    )
                                elif sweeper.type == SweeperType.OFFSET:
                                    sweeper.qs = QbloxSweeper.from_sweeper(program=program, sweeper=sweeper)
                                elif sweeper.type == SweeperType.FACTOR:
                                    raise Exception("SweeperType.FACTOR for Parameter.bias not supported")
                                sweeper.qs.update_parameters = True
                        else:
                            # does not play an active role
                            sweeper.qs = QbloxSweeper(
                                program=program,
                                type=QbloxSweeperType.number,
                                rel_values=range(len(sweeper.values)),
                                name=sweeper.parameter.name,
                            )

                    # FIXME: for qubit sweepers (Parameter.bias, Parameter.attenuation, Parameter.gain), the qubit
                    # information alone is not enough to determine what instrument parameter is to be swept.
                    # For example port gain, both the drive and readout ports have gain parameters.
                    # Until this is resolved, and since bias is only implemented with QCMs offset, this instrument will
                    # never take an active role in those sweeps.

                # Waveforms
                for index, waveform in enumerate(sequencer.waveforms_buffer.unique_waveforms):
                    sequencer.waveforms[waveform.serial] = {"data": waveform.data.tolist(), "index": index}

                # Program
                minimum_delay_between_instructions = 4

                sequence_total_duration = pulses.finish  # the minimum delay between instructions is 4ns
                time_between_repetitions = repetition_duration - sequence_total_duration
                assert time_between_repetitions > minimum_delay_between_instructions
                # TODO: currently relaxation_time needs to be greater than acquisition_hold_off
                # so that the time_between_repetitions is equal to the sequence_total_duration + relaxation_time
                # to be compatible with th erest of the platforms, change it so that time_between_repetitions
                # is equal to pulsesequence duration + acquisition_hold_off if relaxation_time < acquisition_hold_off

                # create registers for key variables
                # nshots is used in the loop that iterates over the number of shots
                nshots_register = Register(program, "nshots")
                # navgs is used in the loop of hardware averages
                navgs_register = Register(program, "navgs")

                header_block = Block("setup")

                body_block = Block()

                body_block.append(f"wait_sync {minimum_delay_between_instructions}")
                if self.ports[port].hardware_mod_en:
                    body_block.append("reset_ph")
                    body_block.append_spacer()

                pulses_block = Block("play")
                # Add an initial wait instruction for the first pulse of the sequence
                initial_wait_block = wait_block(
                    wait_time=pulses.start, register=Register(program), force_multiples_of_four=False
                )
                pulses_block += initial_wait_block

                for n in range(pulses.count):
                    if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.start:
                        pulses_block.append(f"wait {pulses[n].sweeper.register}")

                    if self.ports[port].hardware_mod_en:
                        # # Set frequency
                        # _if = self.get_if(pulses[n])
                        # pulses_block.append(f"set_freq {convert_frequency(_if)}", f"set intermediate frequency to {_if} Hz")

                        # Set phase
                        if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.relative_phase:
                            pulses_block.append(f"set_ph {pulses[n].sweeper.register}")
                        else:
                            pulses_block.append(
                                f"set_ph {convert_phase(pulses[n].relative_phase)}",
                                comment=f"set relative phase {pulses[n].relative_phase} rads",
                            )

                    # Calculate the delay_after_play that is to be used as an argument to the play instruction
                    if len(pulses) > n + 1:
                        # If there are more pulses to be played, the delay is the time between the pulse end and the next pulse start
                        delay_after_play = pulses[n + 1].start - pulses[n].start
                    else:
                        delay_after_play = sequence_total_duration - pulses[n].start

                    if delay_after_play < minimum_delay_between_instructions:
                        raise Exception(
                            f"The minimum delay between the start of two pulses in the same channel is {minimum_delay_between_instructions}ns."
                        )

                    if pulses[n].sweeper and pulses[n].sweeper.type == QbloxSweeperType.duration:
                        RI = pulses[n].sweeper.register
                        if pulses[n].type == PulseType.FLUX:
                            RQ = pulses[n].sweeper.register
                        else:
                            RQ = pulses[n].sweeper.aux_register

                        pulses_block.append(
                            f"play  {RI},{RQ},{delay_after_play}",  # FIXME delay_after_play won't work as the duration increases
                            comment=f"play pulse {pulses[n]} sweeping its duration",
                        )
                    else:
                        # Prepare play instruction: play wave_i_index, wave_q_index, delay_next_instruction
                        pulses_block.append(
                            f"play  {sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_i)},{sequencer.waveforms_buffer.unique_waveforms.index(pulses[n].waveform_q)},{delay_after_play}",
                            comment=f"play waveforms {pulses[n]}",
                        )

                body_block += pulses_block
                body_block.append_spacer()

                final_reset_block = wait_block(
                    wait_time=time_between_repetitions, register=Register(program), force_multiples_of_four=False
                )
                body_block += final_reset_block

                footer_block = Block("cleaup")
                footer_block.append(f"stop")

                # wrap pulses block in sweepers loop blocks
                for sweeper in sweepers:
                    body_block = sweeper.qs.block(inner_block=body_block)

                nshots_block: Block = loop_block(
                    start=0, stop=nshots, step=1, register=nshots_register, block=body_block
                )
                navgs_block = loop_block(start=0, stop=navgs, step=1, register=navgs_register, block=nshots_block)
                program.add_blocks(header_block, navgs_block, footer_block)

                sequencer.program = repr(program)

    def upload(self):
        """Uploads waveforms and programs of all sequencers and arms them in preparation for execution.

        This method should be called after `process_pulse_sequence()`.
        It configures certain parameters of the instrument based on the needs of resources determined
        while processing the pulse sequence.
        """
        # Setup
        for sequencer_number in self._used_sequencers_numbers:
            target = self.device.sequencers[sequencer_number]
            self._set_device_parameter(target, "sync_en", value=True)
            self._set_device_parameter(target, "marker_ovr_en", value=True)  # Default after reboot = False
            self._set_device_parameter(target, "marker_ovr_value", value=15)  # Default after reboot = 0

        for sequencer_number in self._unused_sequencers_numbers:
            target = self.device.sequencers[sequencer_number]
            self._set_device_parameter(target, "sync_en", value=False)
            self._set_device_parameter(target, "marker_ovr_en", value=True)  # Default after reboot = False
            self._set_device_parameter(target, "marker_ovr_value", value=0)  # Default after reboot = 0
            if sequencer_number >= 4:  # Never disconnect default sequencers
                self._set_device_parameter(target, "channel_map_path0_out0_en", value=False)
                self._set_device_parameter(target, "channel_map_path0_out2_en", value=False)
                self._set_device_parameter(target, "channel_map_path1_out1_en", value=False)
                self._set_device_parameter(target, "channel_map_path1_out3_en", value=False)

        # There seems to be a bug in qblox that when any of the mappings between paths and outputs is set,
        # the general offset goes to 0 (eventhough the parameter will still show the right value).
        # Until that is fixed, I'm going to always set the offset just before playing (bypassing the cache):
        self.device.out0_offset(self._device_parameters[self.device.name + "." + "out0_offset"])
        self.device.out1_offset(self._device_parameters[self.device.name + "." + "out1_offset"])
        self.device.out2_offset(self._device_parameters[self.device.name + "." + "out2_offset"])
        self.device.out3_offset(self._device_parameters[self.device.name + "." + "out3_offset"])

        # Upload waveforms and program
        qblox_dict = {}
        sequencer: Sequencer
        for port in self._output_ports_keys:
            for sequencer in self._sequencers[port]:
                # Add sequence program and waveforms to single dictionary
                qblox_dict[sequencer] = {
                    "waveforms": sequencer.waveforms,
                    "weights": sequencer.weights,
                    "acquisitions": sequencer.acquisitions,
                    "program": sequencer.program,
                }

                # Upload dictionary to the device sequencers
                self.device.sequencers[sequencer.number].sequence(qblox_dict[sequencer])

                # DEBUG: QCM Save sequence to file
                if self._debug_folder != "":
                    filename = self._debug_folder + f"Z_{self.name}_sequencer{sequencer.number}_sequence.json"
                    with open(filename, "w", encoding="utf-8") as file:
                        json.dump(qblox_dict[sequencer], file, indent=4)
                        file.write(sequencer.program)

        # Arm sequencers
        for sequencer_number in self._used_sequencers_numbers:
            self.device.arm_sequencer(sequencer_number)

        # DEBUG: QCM Print Readable Snapshot
        # print(self.name)
        # self.device.print_readable_snapshot(update=True)

        # DEBUG: QCM Save Readable Snapshot
        from qibolab.instruments.qblox.debug import print_readable_snapshot

        if self._debug_folder != "":
            filename = self._debug_folder + f"Z_{self.name}_snapshot.json"
            with open(filename, "w", encoding="utf-8") as file:
                print_readable_snapshot(self.device, file, update=True)

    def play_sequence(self):
        """Executes the sequence of instructions."""

        for sequencer_number in self._used_sequencers_numbers:
            # Start used sequencers
            self.device.start_sequencer(sequencer_number)

    def start(self):
        """Empty method to comply with Instrument interface."""

        settings: ClusterQCM_BB_Settings = self.settings
        if self.is_connected:
            try:
                for port in ["o1", "o2", "o3", "o4"]:
                    if port in settings.ports:
                        port_settings: ClusterBB_OutputPort_Settings = settings.ports[port]
                        self.ports[port].offset = port_settings.offset
            except:
                log.warning("Unable to set offsets")

    def stop(self):
        """Stops all sequencers"""

        for sequencer_number in self._used_sequencers_numbers:
            state = self.device.get_sequencer_state(sequencer_number)
            if state.status != "STOPPED":
                log.warning(
                    f"Device {self.device.sequencers[sequencer_number].name} did not stop normally\nstate: {state}"
                )
        try:
            self.device.stop_sequencer()
        except:
            log.warning("Unable to stop sequencers")

        try:
            for port in self.ports:
                self.ports[port].offset = 0

            # self._set_device_parameter(self.device, "out0_offset", value=0)
            # self._set_device_parameter(self.device, "out1_offset", value=0)
            # self._set_device_parameter(self.device, "out2_offset", value=0)
            # self._set_device_parameter(self.device, "out3_offset", value=0)
            # self.device.out0_offset(0)
            # self.device.out1_offset(0)
            # self.device.out2_offset(0)
            # self.device.out3_offset(0)
        except:
            log.warning("Unable to clear offsets")

    def disconnect(self):
        """Empty method to comply with Instrument interface."""
        self._erase_device_parameters_cache()
        self._cluster = None
        self.is_connected = False



from collections import defaultdict
from dataclasses import dataclass, field, fields, replace
from enum import Flag, auto
from typing import List, Optional, Union

from qibo import gates
from qibo.config import raise_error

from qibolab.pulses import FluxPulse, PulseConstructor, PulseSequence, PulseType


class NativeType(Flag):
    """Define available types of native gates.

    Should have the same names with qibo gates.
    """

    M = auto()
    Z = auto()
    RZ = auto()
    GPI2 = auto()
    CZ = auto()
    iSWAP = auto()

    @classmethod
    def from_gate(cls, gate: gates.Gate):
        try:
            return getattr(cls, gate.__class__.__name__)
        except AttributeError:
            raise_error(ValueError, f"Gate {gate} cannot be used as native.")


@dataclass
class NativePulse:
    """Container with parameters required to generate a pulse implementing a native gate."""

    name: str
    """Name of the gate that the pulse implements."""
    duration: int
    amplitude: float
    shape: str
    pulse_type: PulseType
    qubit: "qubits.Qubit"
    frequency: int = 0
    relative_start: int = 0
    """Relative start is relevant for two-qubit gate operations which correspond to a pulse sequence."""

    # used for qblox
    if_frequency: Optional[int] = None
    # TODO: Note sure if the following parameters are useful to be in the runcard
    start: int = 0
    phase: float = 0.0

    @classmethod
    def from_dict(cls, name, pulse, qubit):
        """Parse the dictionary provided by the runcard.

        Args:
            name (str): Name of the native gate (dictionary key).
            pulse (dict): Dictionary containing the parameters of the pulse implementing
                the gate, as loaded from the runcard.
            qubits (:class:`qibolab.platforms.abstract.Qubit`): Qubit that the
                pulse is acting on
        """
        kwargs = pulse.copy()
        kwargs["pulse_type"] = PulseType(kwargs.pop("type"))
        kwargs["qubit"] = qubit
        return cls(name, **kwargs)

    @property
    def raw(self):
        data = {fld.name: getattr(self, fld.name) for fld in fields(self) if getattr(self, fld.name) is not None}
        del data["name"]
        del data["start"]
        if self.pulse_type is PulseType.FLUX:
            del data["frequency"]
            del data["phase"]
        data["qubit"] = self.qubit.name
        data["type"] = data.pop("pulse_type").value
        return data

    def pulse(self, start, relative_phase=0.0):
        """Construct the :class:`qibolab.pulses.Pulse` object implementing the gate.

        Args:
            start (int): Start time of the pulse in the sequence.
            relative_phase (float): Relative phase of the pulse.

        Returns:
            A :class:`qibolab.pulses.DrivePulse` or :class:`qibolab.pulses.DrivePulse`
            or :class:`qibolab.pulses.FluxPulse` with the pulse parameters of the gate.
        """
        if self.pulse_type is PulseType.FLUX:
            return FluxPulse(
                start + self.relative_start,
                self.duration,
                self.amplitude,
                self.shape,
                channel=self.qubit.flux.name,
                qubit=self.qubit.name,
            )

        pulse_cls = PulseConstructor[self.pulse_type.name].value
        channel = getattr(self.qubit, self.pulse_type.name.lower()).name
        return pulse_cls(
            start + self.relative_start,
            self.duration,
            self.amplitude,
            self.frequency,
            relative_phase,
            self.shape,
            channel,
            qubit=self.qubit.name,
        )


@dataclass
class VirtualZPulse:
    """Container with parameters required to add a virtual Z phase in a pulse sequence."""

    phase: float
    qubit: "qubits.Qubit"

    @property
    def raw(self):
        return {"type": "virtual_z", "phase": self.phase, "qubit": self.qubit.name}


@dataclass
class NativeSequence:
    """List of :class:`qibolab.platforms.native.NativePulse` objects implementing a gate.

    Relevant for two-qubit gates, which usually require a sequence of pulses to be implemented.
    These pulses may act on qubits different than the qubits the gate is targeting.
    """

    name: str
    pulses: List[Union[NativePulse, VirtualZPulse]] = field(default_factory=list)

    @classmethod
    def from_dict(cls, name, sequence, qubits):
        """Constructs the native sequence from the dictionaries provided in the runcard.

        Args:
            name (str): Name of the gate the sequence is applying.
            sequence (dict): Dictionary describing the sequence as provided in the runcard.
            qubits (list): List of :class:`qibolab.platforms.abstract.Qubit` object for all
                qubits in the platform. All qubits are required because the sequence may be
                acting on qubits that the implemented gate is not targeting.
        """
        pulses = []

        # If sequence contains only one pulse dictionary, convert it into a list that can be iterated below
        if isinstance(sequence, dict):
            sequence = [sequence]

        for i, pulse in enumerate(sequence):
            pulse = pulse.copy()
            qubit = qubits[pulse.pop("qubit")]
            pulse_type = pulse.pop("type")
            if pulse_type == "virtual_z":
                phase = pulse["phase"]
                pulses.append(VirtualZPulse(phase, qubit))
            else:
                pulses.append(NativePulse(f"{name}{i}", **pulse, pulse_type=PulseType(pulse_type), qubit=qubit))
        return cls(name, pulses)

    @property
    def raw(self):
        return [pulse.raw for pulse in self.pulses]

    def sequence(self, start=0):
        """Creates a :class:`qibolab.pulses.PulseSequence` object implementing the sequence."""
        sequence = PulseSequence()
        virtual_z_phases = defaultdict(int)

        for pulse in self.pulses:
            if isinstance(pulse, NativePulse):
                sequence.add(pulse.pulse(start=start))
            else:
                virtual_z_phases[pulse.qubit.name] += pulse.phase

        # TODO: Maybe ``virtual_z_phases`` should be an attribute of ``PulseSequence``
        return sequence, virtual_z_phases


@dataclass
class SingleQubitNatives:
    """Container with the native single-qubit gates acting on a specific qubit."""

    RX: Optional[NativePulse] = None
    MZ: Optional[NativePulse] = None

    @property
    def RX90(self) -> NativePulse:
        """RX90 native pulse is inferred from RX by halving its amplitude."""
        return replace(self.RX, name="RX90", amplitude=self.RX.amplitude / 2.0)

    @classmethod
    def from_dict(cls, qubit, native_gates):
        """Parse native gates of the qubit from the runcard.

        Args:
            qubit (:class:`qibolab.platforms.abstract.Qubit`): Qubit object that the
                native gates are acting on.
            native_gates (dict): Dictionary with native gate pulse parameters as loaded
                from the runcard.
        """
        pulses = {n: NativePulse.from_dict(n, pulse, qubit=qubit) for n, pulse in native_gates.items()}
        return cls(**pulses)

    @property
    def raw(self):
        """Serialize native gate pulses. ``None`` gates are not included."""
        data = {}
        for fld in fields(self):
            attr = getattr(self, fld.name)
            if attr is not None:
                data[fld.name] = attr.raw
                del data[fld.name]["qubit"]
        return data


@dataclass
class TwoQubitNatives:
    """Container with the native two-qubit gates acting on a specific pair of qubits."""

    CZ: Optional[NativeSequence] = None
    iSWAP: Optional[NativeSequence] = None

    @classmethod
    def from_dict(cls, qubits, native_gates):
        sequences = {n: NativeSequence.from_dict(n, seq, qubits) for n, seq in native_gates.items()}
        return cls(**sequences)

    @property
    def raw(self):
        data = {}
        for fld in fields(self):
            gate = getattr(self, fld.name)
            if gate is not None:
                data[fld.name] = gate.raw
        return data

    @property
    def types(self):
        gate_types = NativeType(0)
        for fld in fields(self):
            gate = fld.name
            if getattr(self, gate) is not None:
                gate_types |= NativeType[gate]
        return gate_types



"""ERAsynth drivers.

Supports the ERAsynth ++.

https://qcodes.github.io/Qcodes_contrib_drivers/_modules/qcodes_contrib_drivers/drivers/ERAInstruments/erasynth.html#ERASynthBase.clear_read_buffer
"""

import json

import requests
from qcodes_contrib_drivers.drivers.ERAInstruments import ERASynthPlusPlus
from qibo.config import log

from qibolab.instruments.abstract import InstrumentException
from qibolab.instruments.oscillator import LocalOscillator

MAX_RECONNECTION_ATTEMPTS = 10
TIMEOUT = 10


class ERA(LocalOscillator):
    def __init__(self, name, address, ethernet=True, reference_clock_source="internal"):
        super().__init__(name, address)
        self.device: ERASynthPlusPlus = None
        self._power: int = None
        self._frequency: int = None
        self.ethernet = ethernet
        self._device_parameters = {}
        self.reference_clock_source = reference_clock_source

    @property
    def frequency(self):
        if self.is_connected:
            if self.ethernet:
                return int(self._get("frequency"))
            return self.device.get("frequency")
        return self._frequency

    @frequency.setter
    def frequency(self, x):
        self._frequency = x
        if self.is_connected:
            self._set_device_parameter("frequency", x)

    @property
    def power(self):
        if self.is_connected:
            if self.ethernet:
                return float(self._get("amplitude"))
            return self.device.get("power")
        return self._power

    @power.setter
    def power(self, x):
        self._power = x
        if self.is_connected:
            self._set_device_parameter("power", x)

    def connect(self):
        """Connects to the instrument using the IP address set in the runcard."""
        if not self.is_connected:
            for attempt in range(3):
                try:
                    if not self.ethernet:
                        self.device = ERASynthPlusPlus(f"{self.name}", f"TCPIP::{self.address}::INSTR")
                    else:
                        self._post("readAll", 1)
                        self._post("readDiagnostic", 0)
                    self.is_connected = True
                    break
                except KeyError as exc:
                    print(f"Unable to connect:\n{str(exc)}\nRetrying...")
                    self.name += "_" + str(attempt)
                except ConnectionError as exc:
                    print(f"Unable to connect:\n{str(exc)}\nRetrying...")
            if not self.is_connected:
                raise InstrumentException(self, f"Unable to connect to {self.name}")
        else:
            raise RuntimeError("There is an open connection to the instrument already")
        # set proper frequency and power if they were changed before connecting
        if self._frequency is not None:
            self._set_device_parameter("frequency", self._frequency)
        if self._power is not None:
            self._set_device_parameter("power", self._power)

    def _set_device_parameter(self, parameter: str, value):
        """Sets a parameter of the instrument, if it changed from the last stored in the cache.

        Args:
            parameter: str = The parameter to be cached and set.
            value = The value to set the paramter.
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        if not (parameter in self._device_parameters and self._device_parameters[parameter] == value):
            if self.is_connected:
                if not self.ethernet:
                    if not hasattr(self.device, parameter):
                        raise ValueError(f"The instrument {self.name} does not have parameter {parameter}")
                    self.device.set(parameter, value)
                else:
                    if parameter == "power":
                        self._post("amplitude", float(value))
                    elif parameter == "frequency":
                        self._post("frequency", int(value))
                self._device_parameters[parameter] = value
            else:
                raise ConnectionError(f"Attempting to set {parameter} without a connection to the instrument")

    def _erase_device_parameters_cache(self):
        """Erases the cache of the instrument parameters."""
        self._device_parameters = {}

    def setup(self, frequency=None, power=None, **kwargs):
        """Configures the instrument.

        A connection to the instrument needs to be established beforehand.

        Args:
            **kwargs: dict = A dictionary of settings loaded from the runcard:
                kwargs["power"]
                kwargs["frequency"]
        Raises:
            Exception = If attempting to set a parameter without a connection to the instrument.
        """
        if frequency is None:
            frequency = self.frequency
        if power is None:
            power = self.power

        if self.is_connected:
            # Load settings
            self.power = power
            self.frequency = frequency

            if not "reference_clock_source" in kwargs:
                kwargs["reference_clock_source"] = self.reference_clock_source
            if not self.ethernet:
                if kwargs["reference_clock_source"] == "internal":
                    self.device.ref_osc_source("int")
                elif kwargs["reference_clock_source"] == "external":
                    self.device.ref_osc_source("ext")
                else:
                    raise ValueError(f"Invalid reference clock source {kwargs['reference_clock_source']}")
            else:
                self._post("rfoutput", 0)

                if kwargs["reference_clock_source"] == "internal":
                    self._post("reference_int_ext", 0)
                elif kwargs["reference_clock_source"] == "external":
                    self._post("reference_int_ext", 1)
                else:
                    raise ValueError(f"Invalid reference clock source {kwargs['reference_clock_source']}")
        else:
            raise ConnectionError("There is no connection to the instrument")

    def start(self):
        self.on()

    def stop(self):
        self.off()

    def disconnect(self):
        if self.is_connected:
            self.is_connected = False

    def __del__(self):
        self.disconnect()

    def on(self):
        if not self.ethernet:
            self.device.on()
        else:
            self._post("rfoutput", 1)

    def off(self):
        if not self.ethernet:
            self.device.off()
        else:
            self._post("rfoutput", 0)

    def close(self):
        self.is_connected = False

    def _post(self, name, value):
        """
        Post a value to the instrument's web server.

        Try to post three times, waiting for 0.1 seconds between each attempt.

        Args:
            name: str = The name of the value to post.
            value: str = The value to post.
        """
        value = str(value)
        for _ in range(MAX_RECONNECTION_ATTEMPTS):
            try:
                response = requests.post(f"http://{self.address}/", data={name: value}, timeout=TIMEOUT)
                if response.status_code == 200:
                    return True
                break
            except (ConnectionError, TimeoutError, requests.exceptions.ReadTimeout):
                log.info("ERAsynth connection timed out, retrying...")
        raise InstrumentException(self, f"Unable to post {name}={value} to {self.name}")

    def _get(self, name):
        """
        Get a value from the instrument's web server.

        Try to get three times, waiting for 0.1 seconds between each attempt.

        Args:
            name: str = The name of the value to get.
        """
        for _ in range(MAX_RECONNECTION_ATTEMPTS):
            try:
                response = requests.post(f"http://{self.address}/", params={"readAll": 1}, timeout=TIMEOUT)

                if response.status_code == 200:
                    # reponse.text is a dictonary in string format, convert it to a dictonary
                    return json.loads(response.text)[name]
                break
            except (ConnectionError, TimeoutError, requests.exceptions.ReadTimeout):
                log.info("ERAsynth connection timed out, retrying...")
        raise InstrumentException(self, f"Unable to get {name} from {self.name}")



from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional

import numpy as np
from qibo.config import log

from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.instruments.abstract import Controller
from qibolab.instruments.port import Port
from qibolab.platform import Qubit
from qibolab.pulses import PulseSequence
from qibolab.qubits import QubitId
from qibolab.sweeper import Sweeper


@dataclass
class DummyPort(Port):
    name: str
    offset: float = 0.0
    lo_frequency: int = 0
    lo_power: int = 0
    gain: int = 0
    attenuation: int = 0
    power_range: int = 0
    filters: Optional[dict] = None


class DummyInstrument(Controller):
    """Dummy instrument that returns random voltage values.

    Useful for testing code without requiring access to hardware.

    Args:
        name (str): name of the instrument.
        address (int): address to connect to the instrument.
            Not used since the instrument is dummy, it only
            exists to keep the same interface with other
            instruments.
    """

    PortType = DummyPort
    sampling_rate = 1

    def connect(self):
        log.info("Connecting to dummy instrument.")

    def setup(self, *args, **kwargs):
        log.info("Setting up dummy instrument.")

    def start(self):
        log.info("Starting dummy instrument.")

    def stop(self):
        log.info("Stopping dummy instrument.")

    def disconnect(self):
        log.info("Disconnecting dummy instrument.")

    def get_values(self, options, sequence, shape):
        for ro_pulse in sequence.ro_pulses:
            if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                if options.averaging_mode is AveragingMode.SINGLESHOT:
                    values = np.random.randint(2, size=shape)
                elif options.averaging_mode is AveragingMode.CYCLIC:
                    values = np.random.rand(*shape)
            elif options.acquisition_type is AcquisitionType.RAW:
                samples = int(ro_pulse.duration * self.sampling_rate)
                waveform_shape = tuple(samples * dim for dim in shape)
                values = np.random.rand(*waveform_shape) * 100 + 1j * np.random.rand(*waveform_shape) * 100
            elif options.acquisition_type is AcquisitionType.INTEGRATION:
                values = np.random.rand(*shape) * 100 + 1j * np.random.rand(*shape) * 100
        return values

    def play(
        self,
        qubits: Dict[QubitId, Qubit],
        sequence: PulseSequence,
        options: ExecutionParameters,
    ):
        exp_points = 1 if options.averaging_mode is AveragingMode.CYCLIC else options.nshots
        shape = (exp_points,)
        results = {}

        for ro_pulse in sequence.ro_pulses:
            values = self.get_values(options, sequence, shape)
            results[ro_pulse.qubit] = results[ro_pulse.serial] = options.results_type(values)

        return results

    def play_sequences(
        self,
        qubits: Dict[QubitId, Qubit],
        sequences: List[PulseSequence],
        options: ExecutionParameters,
    ):
        exp_points = 1 if options.averaging_mode is AveragingMode.CYCLIC else options.nshots
        shape = (exp_points,)

        results = defaultdict(list)
        for sequence in sequences:
            for ro_pulse in sequence.ro_pulses:
                values = self.get_values(options, sequence, shape)
                results[ro_pulse.serial].append(options.results_type(values))
                results[ro_pulse.qubit].append(options.results_type(values))

        return results

    def sweep(
        self,
        qubits: Dict[QubitId, Qubit],
        sequence: PulseSequence,
        options: ExecutionParameters,
        *sweepers: List[Sweeper],
    ):
        results = {}

        if options.averaging_mode is not AveragingMode.CYCLIC:
            shape = (options.nshots,) + tuple(len(sweeper.values) for sweeper in sweepers)
        else:
            shape = tuple(len(sweeper.values) for sweeper in sweepers)

        for ro_pulse in sequence.ro_pulses:
            values = self.get_values(options, sequence, shape)
            results[ro_pulse.qubit] = results[ro_pulse.serial] = options.results_type(values)

        return results



from dataclasses import dataclass

from qm import SimulationConfig
from qm.QuantumMachinesManager import QuantumMachinesManager
from qualang_tools.simulator_tools import create_simulator_controller_connections

from .driver import QMOPX


@dataclass
class QMSim(QMOPX):
    """Instrument for using the Quantum Machines (QM) OPX simulator.

    Args:
        address (str): Address and port of the simulator.
        simulation_duration (int): Duration for the simulation in ns.
        cloud (bool): If ``True`` the QM cloud simulator is used which does not
            require access to physical instruments. This assumes that a proper
            cloud address has been given.
            If ``False`` the simulator built-in the instruments is used.
            This requires connection to instruments.
            Default is ``False``.
    """

    simulation_duration: int = 1000
    cloud: bool = False

    def __post_init__(self):
        """Convert simulation duration from ns to clock cycles."""
        self.simulation_duration //= 4

    def connect(self):
        host, port = self.address.split(":")
        if self.cloud:
            from qm.simulate.credentials import create_credentials

            self.manager = QuantumMachinesManager(host, int(port), credentials=create_credentials())
        else:
            self.manager = QuantumMachinesManager(host, int(port))

    @staticmethod
    def fetch_results(result, ro_pulses):
        return result

    def execute_program(self, program):
        ncontrollers = len(self.config.controllers)
        controller_connections = create_simulator_controller_connections(ncontrollers)
        simulation_config = SimulationConfig(
            duration=self.simulation_duration, controller_connections=controller_connections
        )
        return self.manager.simulate(self.config.__dict__, program, simulation_config)



import collections
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Union

import numpy as np
from numpy import typing as npt
from qibo.config import raise_error
from qm import qua
from qm.qua._dsl import _Variable  # for type declaration only
from qualang_tools.bakery import baking
from qualang_tools.bakery.bakery import Baking

from qibolab import AcquisitionType, AveragingMode
from qibolab.instruments.qm.acquisition import (
    Acquisition,
    IntegratedAcquisition,
    RawAcquisition,
    ShotsAcquisition,
)
from qibolab.pulses import Pulse, PulseType
from qibolab.sweeper import Parameter

from .config import QMConfig

DurationsType = Union[List[int], npt.NDArray[int]]
"""Type of values that can be accepted in a duration sweeper."""


@dataclass
class QMPulse:
    """Wrapper around :class:`qibolab.pulses.Pulse` for easier translation to QUA program.

    These pulses are defined when :meth:`qibolab.instruments.qm.QMOPX.play` is called
    and hold attributes for the ``element`` and ``operation`` that corresponds to each pulse,
    as defined in the QM config.
    """

    pulse: Pulse
    """:class:`qibolab.pulses.Pulse` implemting the current pulse."""
    element: Optional[str] = None
    """Element that the pulse will be played on, as defined in the QM config."""
    operation: Optional[str] = None
    """Name of the operation that is implementing the pulse in the QM config."""
    relative_phase: Optional[float] = None
    """Relative phase of the pulse normalized to follow QM convention.
    May be overwritten when sweeping phase."""
    wait_time: int = 0
    """Time (in clock cycles) to wait before playing this pulse.
    Calculated and assigned by :meth:`qibolab.instruments.qm.Sequence.add`."""
    wait_time_variable: Optional[_Variable] = None
    """Time (in clock cycles) to wait before playing this pulse when we are sweeping start."""
    swept_duration: Optional[_Variable] = None
    """Pulse duration when sweeping it."""

    acquisition: Optional[Acquisition] = None
    """Data class containing the variables required for data acquisition for the instrument."""

    next_: Set["QMPulse"] = field(default_factory=set)
    """Pulses that will be played after the current pulse.
    These pulses need to be re-aligned if we are sweeping the start or duration."""
    elements_to_align: Set[str] = field(default_factory=set)

    def __post_init__(self):
        self.element: str = f"{self.pulse.type.name.lower()}{self.pulse.qubit}"
        self.operation: str = self.pulse.serial
        self.relative_phase: float = self.pulse.relative_phase / (2 * np.pi)
        self.elements_to_align.add(self.element)

    def __hash__(self):
        return hash(self.pulse)

    @property
    def duration(self):
        """Duration of the pulse as defined in the :class:`qibolab.pulses.PulseSequence`.

        Remains constant even when we are sweeping the duration of this pulse.
        """
        return self.pulse.duration

    @property
    def wait_cycles(self):
        """Instrument clock cycles (1 cycle = 4ns) to wait before playing the pulse.

        This property will be used in the QUA ``wait`` command, so that it is compatible
        with and without start sweepers.
        """
        if self.wait_time_variable is not None:
            return self.wait_time_variable + self.wait_time
        if self.wait_time >= 4:
            return self.wait_time
        return None

    def play(self):
        """Play the pulse.

        Relevant only in the context of a QUA program.
        """
        qua.play(self.operation, self.element, duration=self.swept_duration)

    def declare_output(self, options, threshold=None, angle=None):
        average = options.averaging_mode is AveragingMode.CYCLIC
        acquisition_type = options.acquisition_type
        if acquisition_type is AcquisitionType.RAW:
            self.acquisition = RawAcquisition(self.pulse.serial, average)
        elif acquisition_type is AcquisitionType.INTEGRATION:
            self.acquisition = IntegratedAcquisition(self.pulse.serial, average)
        elif acquisition_type is AcquisitionType.DISCRIMINATION:
            if threshold is None or angle is None:
                raise_error(
                    ValueError, "Cannot use ``AcquisitionType.DISCRIMINATION`` " "if threshold and angle are not given."
                )
            self.acquisition = ShotsAcquisition(self.pulse.serial, average, threshold, angle)
        else:
            raise_error(ValueError, f"Invalid acquisition type {acquisition_type}.")
        self.acquisition.assign_element(self.element)


@dataclass
class BakedPulse(QMPulse):
    """Baking allows 1ns resolution in the pulse waveforms."""

    segments: List[Baking] = field(default_factory=list)
    """Baked segments implementing the pulse."""
    amplitude: Optional[float] = None
    """Amplitude of the baked pulse. Relevant only when sweeping amplitude."""
    durations: Optional[DurationsType] = None

    def __hash__(self):
        return super().__hash__()

    @property
    def duration(self):
        return self.segments[-1].get_op_length()

    @staticmethod
    def calculate_waveform(original_waveform, t):
        if t == 0:  # Otherwise, the baking will be empty and will not be created
            return [0.0] * 16

        expanded_waveform = list(original_waveform)
        for i in range(t // len(original_waveform)):
            expanded_waveform.extend(original_waveform)
        return expanded_waveform[:t]

    def bake(self, config: QMConfig, durations: DurationsType):
        self.segments = []
        self.durations = durations
        for t in durations:
            with baking(config.__dict__, padding_method="right") as segment:
                if self.pulse.type is PulseType.FLUX:
                    waveform = self.pulse.envelope_waveform_i.data.tolist()
                    waveform = self.calculate_waveform(waveform, t)
                else:
                    waveform_i = self.pulse.envelope_waveform_i.data.tolist()
                    waveform_q = self.pulse.envelope_waveform_q.data.tolist()
                    waveform = [
                        self.calculate_waveform(waveform_i, t),
                        self.calculate_waveform(waveform_q, t),
                    ]
                segment.add_op(self.pulse.serial, self.element, waveform)
                segment.play(self.pulse.serial, self.element)
            self.segments.append(segment)

    @property
    def amplitude_array(self):
        if self.amplitude is None:
            return None
        return [(self.element, self.amplitude)]

    def play(self):
        if self.swept_duration is not None:
            with qua.switch_(self.swept_duration):
                for dur, segment in zip(self.durations, self.segments):
                    with qua.case_(dur):
                        segment.run(amp_array=self.amplitude_array)
        else:
            segment = self.segments[0]
            segment.run(amp_array=self.amplitude_array)


def find_duration_sweeper_pulses(sweepers):
    """Find all pulses that require baking because we are sweeping their duration."""
    duration_sweep_pulses = set()
    for sweeper in sweepers:
        try:
            step = sweeper.values[1] - sweeper.values[0]
        except IndexError:
            step = sweeper.values[0]

        if sweeper.parameter is Parameter.duration and step % 4 != 0:
            for pulse in sweeper.pulses:
                duration_sweep_pulses.add(pulse.serial)

    return duration_sweep_pulses


@dataclass
class Sequence:
    """Pulse sequence containing QM specific pulses (``qmpulse``).
    Defined in :meth:`qibolab.instruments.qm.QMOPX.play`.
    Holds attributes for the ``element`` and ``operation`` that
    corresponds to each pulse, as defined in the QM config.
    """

    qmpulses: List[QMPulse] = field(default_factory=list)
    """List of :class:`qibolab.instruments.qm.QMPulse` objects corresponding to the original pulses."""
    ro_pulses: List[QMPulse] = field(default_factory=list)
    """List of readout pulses used for registering outputs."""
    pulse_to_qmpulse: Dict[Pulse, QMPulse] = field(default_factory=dict)
    """Map from qibolab pulses to QMPulses (useful when sweeping)."""
    clock: Dict[str, int] = field(default_factory=lambda: collections.defaultdict(int))
    """Dictionary used to keep track of times of each element, in order to calculate wait times."""
    pulse_finish: Dict[int, List[QMPulse]] = field(default_factory=lambda: collections.defaultdict(list))
    """Map to find all pulses that finish at a given time (useful for ``_find_previous``)."""

    @classmethod
    def create(cls, qubits, sequence, sweepers, config, time_of_flight, smearing):
        """Translates a :class:`qibolab.pulses.PulseSequence` to a :class:`qibolab.instruments.qm.sequence.Sequence`.

        Args:
            qubits (list): List of :class:`qibolab.platforms.abstract.Qubit` objects
                passed from the platform.
            sequence (:class:`qibolab.pulses.PulseSequence`). Pulse sequence to translate.
        Returns:
            (:class:`qibolab.instruments.qm.Sequence`) containing the pulses from given pulse sequence.
        """
        # Current driver cannot play overlapping pulses on drive and flux channels
        # If we want to play overlapping pulses we need to define different elements on the same ports
        # like we do for readout multiplex
        duration_sweep_pulses = find_duration_sweeper_pulses(sweepers)
        qmsequence = cls()
        for pulse in sorted(sequence.pulses, key=lambda pulse: (pulse.start, pulse.duration)):
            config.register_element(qubits[pulse.qubit], pulse, time_of_flight, smearing)
            if pulse.duration % 4 != 0 or pulse.duration < 16 or pulse.serial in duration_sweep_pulses:
                qmpulse = BakedPulse(pulse)
                qmpulse.bake(config, durations=[pulse.duration])
            else:
                qmpulse = QMPulse(pulse)
                config.register_pulse(qubits[pulse.qubit], pulse)
            qmsequence.add(qmpulse)

        qmsequence.shift()

        return qmsequence

    def _find_previous(self, pulse):
        for finish in reversed(sorted(self.pulse_finish.keys())):
            if finish <= pulse.start:
                # first try to find a previous pulse targeting the same qubit
                last_pulses = self.pulse_finish[finish]
                for previous in reversed(last_pulses):
                    if previous.pulse.qubit == pulse.qubit:
                        return previous
                # otherwise
                if finish == pulse.start:
                    return last_pulses[-1]
        return None

    def add(self, qmpulse: QMPulse):
        pulse = qmpulse.pulse
        self.pulse_to_qmpulse[pulse.serial] = qmpulse
        if pulse.type is PulseType.READOUT:
            self.ro_pulses.append(qmpulse)

        previous = self._find_previous(pulse)
        if previous is not None:
            previous.next_.add(qmpulse)

        wait_time = pulse.start - self.clock[qmpulse.element]
        if wait_time >= 12:
            qmpulse.wait_time = wait_time // 4 + 1
            self.clock[qmpulse.element] += 4 * qmpulse.wait_time
        self.clock[qmpulse.element] += qmpulse.duration

        self.pulse_finish[pulse.finish].append(qmpulse)
        self.qmpulses.append(qmpulse)

    def shift(self):
        """Shift all pulses that come after a ``BakedPulse`` a bit to avoid overlapping pulses."""
        to_shift = collections.deque()
        for qmpulse in self.qmpulses:
            if isinstance(qmpulse, BakedPulse):
                to_shift.extend(qmpulse.next_)
        while to_shift:
            qmpulse = to_shift.popleft()
            qmpulse.wait_time += 2
            to_shift.extend(qmpulse.next_)

    def play(self, relaxation_time=0):
        """Part of QUA program that plays an arbitrary pulse sequence.

        Should be used inside a ``program()`` context.
        """
        needs_reset = False
        qua.align()
        for qmpulse in self.qmpulses:
            pulse = qmpulse.pulse
            if qmpulse.wait_cycles is not None:
                qua.wait(qmpulse.wait_cycles, qmpulse.element)
            if pulse.type is PulseType.READOUT:
                qmpulse.acquisition.measure(qmpulse.operation, qmpulse.element)
            else:
                if not isinstance(qmpulse.relative_phase, float) or qmpulse.relative_phase != 0:
                    qua.frame_rotation_2pi(qmpulse.relative_phase, qmpulse.element)
                    needs_reset = True
                qmpulse.play()
                if needs_reset:
                    qua.reset_frame(qmpulse.element)
                    needs_reset = False
                if len(qmpulse.elements_to_align) > 1:
                    qua.align(*qmpulse.elements_to_align)

        if relaxation_time > 0:
            qua.wait(relaxation_time // 4)

        # Save data to the stream processing
        for qmpulse in self.ro_pulses:
            qmpulse.acquisition.save()



from functools import cached_property, lru_cache
from typing import Optional

import numpy as np
import numpy.typing as npt


class IntegratedResults:
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.INTEGRATION and AveragingMode.SINGLESHOT
    """

    def __init__(self, data: np.ndarray):
        self.voltage: npt.NDArray[np.complex128] = data

    def __add__(self, data):
        return self.__class__(np.append(self.voltage, data.voltage))

    @property
    def voltage_i(self):
        """Signal component i in volts."""
        return self.voltage.real

    @property
    def voltage_q(self):
        """Signal component q in volts."""
        return self.voltage.imag

    @cached_property
    def magnitude(self):
        """Signal magnitude in volts."""
        return np.sqrt(self.voltage_i**2 + self.voltage_q**2)

    @cached_property
    def phase(self):
        """Signal phase in radians."""
        return np.angle(self.voltage_i + 1.0j * self.voltage_q)

    @property
    def serialize(self):
        """Serialize as a dictionary."""
        serialized_dict = {
            "MSR[V]": self.magnitude.flatten(),
            "i[V]": self.voltage_i.flatten(),
            "q[V]": self.voltage_q.flatten(),
            "phase[rad]": self.phase.flatten(),
        }
        return serialized_dict

    @property
    def average(self):
        """Perform average over i and q"""
        average_data = np.mean(self.voltage, axis=0)
        std_data = np.std(self.voltage, axis=0, ddof=1) / np.sqrt(self.voltage.shape[0])
        return AveragedIntegratedResults(average_data, std_data)


class AveragedIntegratedResults(IntegratedResults):
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.INTEGRATION and AveragingMode.CYCLIC
    or the averages of ``IntegratedResults``
    """

    def __init__(self, data: np.ndarray, std: np.ndarray = np.array([])):
        super().__init__(data)
        self.std: Optional[npt.NDArray[np.float64]] = std

    def __add__(self, data):
        new_res = super().__add__(data)
        new_res.std = np.append(self.std, data.std)
        return new_res


class RawWaveformResults(IntegratedResults):
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.RAW and AveragingMode.SINGLESHOT
    may also be used to store the integration weights ?
    """


class AveragedRawWaveformResults(AveragedIntegratedResults):
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.RAW and AveragingMode.CYCLIC
    or the averages of ``RawWaveformResults``
    """


class SampleResults:
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.DISCRIMINATION and AveragingMode.SINGLESHOT
    """

    def __init__(self, data: np.ndarray):
        self.samples: npt.NDArray[np.uint32] = data

    def __add__(self, data):
        return self.__class__(np.append(self.samples, data.samples))

    @lru_cache
    def probability(self, state=0):
        """Returns the statistical frequency of the specified state (0 or 1)."""
        return abs(1 - state - np.mean(self.samples, axis=0))

    @property
    def serialize(self):
        """Serialize as a dictionary."""
        serialized_dict = {
            "0": self.probability(0).flatten(),
        }
        return serialized_dict

    @property
    def average(self):
        """Perform samples average"""
        average = self.probability(1)
        std = np.std(self.samples, axis=0, ddof=1) / np.sqrt(self.samples.shape[0])
        return AveragedSampleResults(average, self.samples, std=std)


class AveragedSampleResults(SampleResults):
    """
    Data structure to deal with the output of
    :func:`qibolab.platforms.abstract.AbstractPlatform.execute_pulse_sequence`
    :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`

    Associated with AcquisitionType.DISCRIMINATION and AveragingMode.CYCLIC
    or the averages of ``SampleResults``
    """

    def __init__(
        self, statistical_frequency: np.ndarray, samples: np.ndarray = np.array([]), std: np.ndarray = np.array([])
    ):
        super().__init__(samples)
        self.statistical_frequency: npt.NDArray[np.float64] = statistical_frequency
        self.std: Optional[npt.NDArray[np.float64]] = std

    def __add__(self, data):
        new_res = super().__add__(data)
        new_res.statistical_frequency = np.append(self.statistical_frequency, data.statistical_frequency)
        new_res.std = np.append(self.std, data.std)
        return new_res

    @lru_cache
    def probability(self, state=0):
        """Returns the statistical frequency of the specified state (0 or 1)."""
        return abs(1 - state - self.statistical_frequency)



import bisect
import socket
import struct
import threading
import time
from typing import List

import numpy as np

from qibolab.instruments.abstract import Instrument
from qibolab.pulses import Pulse


class PulseBlaster(Instrument):
    """Driver for the 24-pin PulseBlaster TTL signal generator."""

    def __init__(self, name, address, port=5000):
        super().__init__(name, address)
        self.port = port
        self._pins = None

    def setup(self, holdtime_ns, pins=None, **kwargs):
        """Setup the PulseBlaster.

        Arguments:
            holdtime_ns (int): TTL pulse length and delay between TTL pulses. The experiment repetition frequency is 1 / (2 * holdtime_ns).
            pins (int): Pins to trigger in hex, defaults to all pins.
        """
        if pins is None:
            pins = list(range(24))
        self._pins = self._hexify(pins)
        self._holdtime = holdtime_ns

    def arm(self, nshots, readout_start=0):
        """Arm the PulseBlaster for playback. Sends a signal to the instrument to setup the pulse sequence and repetition.

        Arguments:
            nshots (int): Number of TTL triggers to repeat.
        """
        payload = f"setup,{self._pins},{nshots},{self._holdtime}"
        return self._send(payload.encode("utf-8"), True)

    def fire(self):
        """Starts the PulseBlaster."""
        self._send(b"fire")

    def start(self):
        pass

    def stop(self):
        self._send(b"stop")

    def status(self):
        return self._send(b"status", True)

    def _send(self, payload, retval=False):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((self.address, self.port))
            s.sendall(payload)

            if retval:
                return s.recv(1024).decode("utf-8")

    def play_sequence(self):
        self.fire()

    def connect(self):
        pass

    def disconnect(self):
        pass

    @staticmethod
    def _hexify(pins):
        return int("".join(["1" if i in set(pins) else "0" for i in reversed(range(24))]), 2)


class IcarusQFPGA(Instrument):
    """Driver for the IcarusQ RFSoC socket-based implementation."""

    def __init__(self, name, address, port=8080):
        super().__init__(name, address)
        self._dac_sample_size = 65536
        self._adc_sample_size = 65536
        self._dac_nchannels = 16
        self._adc_nchannels = 16
        self._dac_sampling_rate = 5898240000
        self._adc_sampling_rate = 1966080000
        self._nshots = 0

        self.port = port
        self._thread = None
        self._buffer = None
        self._adcs_to_read = None
        self.nshots = None

    def setup(self, dac_sampling_rate, adcs_to_read, **kwargs):
        """Sets the sampling rate of the RFSoC. May need to be repeated several times due to multi-tile sync error.

        Arguments:
            dac_sampling_rate_id (int): Sampling rate ID to be set on the RFSoC.
            dac_sampling_rate_6g_id (int): Optional sampling rate ID for the 6GS/s mode if used.
        """
        self._adcs_to_read = adcs_to_read
        self._dac_sampling_rate = dac_sampling_rate

    def translate(self, sequence: List[Pulse], nshots):
        """Translates the pulse sequence into a numpy array."""

        # Waveform is 14-bit resolution on the DACs, so we first create 16-bit arrays to store the data.
        waveform = np.zeros((self._dac_nchannels, self._dac_sample_size), dtype="i2")

        # The global time can first be set as float to handle rounding errors.
        time_array = 1 / self._dac_sampling_rate * np.arange(0, self._dac_sample_size, 1)

        for pulse in sequence:
            # Get array indices corresponding to the start and end of the pulse. Note that the pulse time parameters are in ns and require conversion.
            start = bisect.bisect(time_array, pulse.start * 1e-9)
            end = bisect.bisect(time_array, (pulse.start + pulse.duration) * 1e-9)

            # Create the pulse waveform and cast it to 16-bit. The ampltiude is max signed 14-bit (+- 8191) and the indices should take care of any overlap of pulses.
            # 2-byte bit shift for downsampling from 16 bit to 14 bit
            pulse_waveform = (4 * np.sin(2 * np.pi * pulse.frequency * time_array[start:end] + pulse.phase)).astype(
                "i2"
            )
            waveform[pulse.channel, start:end] += pulse_waveform

        self.nshots = nshots

        return waveform

    def upload(self, waveform):
        """Uploads a numpy array of size DAC_CHANNELS X DAC_SAMPLE_SIZE to the PL memory.

        Arguments:
            waveform (numpy.ndarray): Numpy array of size DAC_CHANNELS X DAC_SAMPLE_SIZE with type signed short.
        """

        # TODO: Implement checks for size and dtype of waveform.

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((self.address, self.port))
            # Signal to the RFSoC to start listening for DAC waveforms and load into PL mem.
            s.sendall(struct.pack("B", 1))
            s.sendall(waveform.tobytes())

    def play_sequence(self):
        """DACs are automatically armed for playbacked when waveforms are loaded, no need to signal"""
        self._buffer = np.zeros((self._adc_nchannels, self._adc_sample_size))
        self._thread = threading.Thread(target=self._play, args=(self.nshots,))
        self._thread.start()

        time.sleep(0.1)  # Use threading lock and socket signals instead of hard sleep?

    def play_sequence_and_acquire(self):
        """Signal the RFSoC to arm the ADC and start data transfer into PS memory.
        Starts a thread to listen for ADC data from the RFSoC.

        Arguments:
            nshots (int): Number of shots.
        """
        # Create buffer to hold ADC data.
        # TODO: Create flag to handle single shot measurement / buffer assignment per shot.

    def _play(self, nshots):
        """Starts ADC data acquisition and transfer on the RFSoC."""

        if len(self._adcs_to_read) == 0:
            return

        # 2 bytes per ADC data point
        BUF_SIZE = int(self._adc_sample_size * 2)

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((self.address, self.port))
            # Signal RFSoC to arm ADC and expect `nshots` number of triggers.
            s.sendall(struct.pack("B", 2))
            s.sendall(struct.pack("H", nshots))
            s.sendall(struct.pack("B", len(self._adcs_to_read)))  # send number of channels

            for adc in self._adcs_to_read:
                s.sendall(struct.pack("B", adc))  # send ADC channel to read

            # Use the same socket to start listening for ADC data transfer.
            for _ in range(nshots):
                for _ in self._adcs_to_read:
                    # IcarusQ board may send channel/shot data out of order due to threading implementation.
                    # shotnum = struct.unpack("H", s.recv(2))[0]
                    channel = struct.unpack("H", s.recv(2))[0]

                    # Start listening for ADC_SAMPLE_SIZE * 2 bytes corresponding to data per channel.
                    r = bytearray()
                    while len(r) < BUF_SIZE:
                        # Socket implementation does not return exactly desired amount of bytes, keep querying until bytearray reaches expected amount of bytes.
                        # TODO: Look for `MSG_WAITALL` flag in socket recv.
                        packet = s.recv(BUF_SIZE - len(r))
                        if packet:
                            r.extend(packet)

                    # Accumulate ADC data in buffer, buffer float dtype should be enough to prevent overflow.
                    self._buffer[channel] += np.frombuffer(r, dtype="i2")

        # Average buffer at the end of measurement
        self._buffer = self._buffer / nshots

    def result(self, readout_frequency, readout_channel):
        """Returns the processed signal result from the ADC.

        Arguments:
            readout_frequency (float): Frequency to be used for signal processing.
            readout_channel (int): Channel to be used for signal processing.

        Returns:
            ampl (float): Amplitude of the processed signal.
            phase (float): Phase shift of the processed signal in degrees.
            it (float): I component of the processed signal.
            qt (float): Q component of the processed signal.
        """

        # Wait for ADC data acquisition to complete
        self._thread.join()

        input_vec = self._buffer[readout_channel]

        time_vec = 1 / self._adc_sampling_rate * np.arange(0, self._adc_sample_size, 1)
        vec_I = np.sin(2 * np.pi * readout_frequency * time_vec)
        vec_Q = np.cos(2 * np.pi * readout_frequency * time_vec)

        it = np.sum(vec_I * input_vec)
        qt = np.sum(vec_Q * input_vec)
        phase = np.arctan2(qt, it)
        ampl = np.sqrt(it**2 + qt**2)

        return ampl, phase, it, qt

    def start(self):
        pass

    def connect(self):
        pass

    def stop(self):
        pass

    def disconnect(self):
        pass



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np
from qm import qua
from qm.qua import declare, declare_stream, fixed
from qm.qua._dsl import _ResultSource, _Variable  # for type declaration only
from qualang_tools.addons.variables import assign_variables_to_element
from qualang_tools.units import unit

from qibolab.result import (
    AveragedIntegratedResults,
    AveragedRawWaveformResults,
    AveragedSampleResults,
    IntegratedResults,
    RawWaveformResults,
    SampleResults,
)


@dataclass
class Acquisition(ABC):
    """QUA variables used for saving of acquisition results.

    This class can be instantiated only within a QUA program scope.
    Each readout pulse is associated with its own set of acquisition variables.
    """

    serial: str
    """Serial of the readout pulse that generates this acquisition."""
    average: bool

    @abstractmethod
    def assign_element(self, element):
        """Assign acquisition variables to the corresponding QM controlled.

        Proposed to do by QM to avoid crashes.

        Args:
            element (str): Element (from ``config``) that the pulse will be applied on.
        """

    @abstractmethod
    def measure(self, operation, element):
        """Send measurement pulse and acquire results.

        Args:
            operation (str): Operation (from ``config``) corresponding to the pulse to be played.
            element (str): Element (from ``config``) that the pulse will be applied on.
        """

    @abstractmethod
    def save(self):
        """Save acquired results from variables to streams."""

    @abstractmethod
    def download(self, *dimensions):
        """Save streams to prepare for fetching from host device.

        Args:
            dimensions (int): Dimensions to use for buffer of data.
        """

    @abstractmethod
    def fetch(self):
        """Fetch downloaded streams to host device."""


@dataclass
class RawAcquisition(Acquisition):
    """QUA variables used for raw waveform acquisition."""

    adc_stream: _ResultSource = field(default_factory=lambda: declare_stream(adc_trace=True))
    """Stream to collect raw ADC data."""

    def assign_element(self, element):
        pass

    def measure(self, operation, element):
        qua.measure(operation, element, self.adc_stream)

    def save(self):
        pass

    def download(self, *dimensions):
        i_stream = self.adc_stream.input1()
        q_stream = self.adc_stream.input2()
        if self.average:
            i_stream = i_stream.average()
            q_stream = q_stream.average()
        i_stream.save(f"{self.serial}_I")
        q_stream.save(f"{self.serial}_Q")

    def fetch(self, handles):
        ires = handles.get(f"{self.serial}_I").fetch_all()
        qres = handles.get(f"{self.serial}_Q").fetch_all()
        # convert raw ADC signal to volts
        u = unit()
        ires = u.raw2volts(ires)
        qres = u.raw2volts(qres)
        if self.average:
            return AveragedRawWaveformResults(ires + 1j * qres)
        return RawWaveformResults(ires + 1j * qres)


@dataclass
class IntegratedAcquisition(Acquisition):
    """QUA variables used for integrated acquisition."""

    I: _Variable = field(default_factory=lambda: declare(fixed))
    Q: _Variable = field(default_factory=lambda: declare(fixed))
    """Variables to save the (I, Q) values acquired from a single shot."""
    I_stream: _ResultSource = field(default_factory=lambda: declare_stream())
    Q_stream: _ResultSource = field(default_factory=lambda: declare_stream())
    """Streams to collect the results of all shots."""

    def assign_element(self, element):
        assign_variables_to_element(element, self.I, self.Q)

    def measure(self, operation, element):
        qua.measure(
            operation,
            element,
            None,
            qua.dual_demod.full("cos", "out1", "sin", "out2", self.I),
            qua.dual_demod.full("minus_sin", "out1", "cos", "out2", self.Q),
        )

    def save(self):
        qua.save(self.I, self.I_stream)
        qua.save(self.Q, self.Q_stream)

    def download(self, *dimensions):
        Istream = self.I_stream
        Qstream = self.Q_stream
        for dim in dimensions:
            Istream = Istream.buffer(dim)
            Qstream = Qstream.buffer(dim)
        if self.average:
            Istream = Istream.average()
            Qstream = Qstream.average()
        Istream.save(f"{self.serial}_I")
        Qstream.save(f"{self.serial}_Q")

    def fetch(self, handles):
        ires = handles.get(f"{self.serial}_I").fetch_all()
        qres = handles.get(f"{self.serial}_Q").fetch_all()
        if self.average:
            # TODO: calculate std
            return AveragedIntegratedResults(ires + 1j * qres)
        return IntegratedResults(ires + 1j * qres)


@dataclass
class ShotsAcquisition(Acquisition):
    """QUA variables used for shot classification.

    Threshold and angle must be given in order to classify shots.
    """

    threshold: float
    """Threshold to be used for classification of single shots."""
    angle: float
    """Angle in the IQ plane to be used for classification of single shots."""

    I: _Variable = field(default_factory=lambda: declare(fixed))
    Q: _Variable = field(default_factory=lambda: declare(fixed))
    """Variables to save the (I, Q) values acquired from a single shot."""
    shot: _Variable = field(default_factory=lambda: declare(int))
    """Variable for calculating an individual shots."""
    shots: _ResultSource = field(default_factory=lambda: declare_stream())
    """Stream to collect multiple shots."""

    def __post_init__(self):
        self.cos = np.cos(self.angle)
        self.sin = np.sin(self.angle)

    def assign_element(self, element):
        assign_variables_to_element(element, self.I, self.Q, self.shot)

    def measure(self, operation, element):
        qua.measure(
            operation,
            element,
            None,
            qua.dual_demod.full("cos", "out1", "sin", "out2", self.I),
            qua.dual_demod.full("minus_sin", "out1", "cos", "out2", self.Q),
        )
        qua.assign(self.shot, qua.Cast.to_int(self.I * self.cos - self.Q * self.sin > self.threshold))

    def save(self):
        qua.save(self.shot, self.shots)

    def download(self, *dimensions):
        shots = self.shots
        for dim in dimensions:
            shots = shots.buffer(dim)
        if self.average:
            shots = shots.average()
        shots.save(f"{self.serial}_shots")

    def fetch(self, handles):
        shots = handles.get(f"{self.serial}_shots").fetch_all()
        if self.average:
            # TODO: calculate std
            return AveragedSampleResults(shots)
        return SampleResults(shots.astype(int))



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from qibolab.result import (
    AveragedIntegratedResults,
    AveragedRawWaveformResults,
    AveragedSampleResults,
    IntegratedResults,
    RawWaveformResults,
    SampleResults,
)


class AcquisitionType(Enum):
    """Data acquisition from hardware"""

    DISCRIMINATION = auto()
    """Demodulate, integrate the waveform and discriminate among states based on the voltages"""
    INTEGRATION = auto()
    """Demodulate and integrate the waveform"""
    RAW = auto()
    """Acquire the waveform as it is"""
    SPECTROSCOPY = auto()
    """Zurich Integration mode for RO frequency sweeps"""


class AveragingMode(Enum):
    """Data averaging modes from hardware"""

    CYCLIC = auto()
    """Better averaging for short timescale noise"""
    SINGLESHOT = auto()
    """SINGLESHOT: No averaging"""
    SEQUENTIAL = auto()
    """SEQUENTIAL: Worse averaging for noise[Avoid]"""


RESULTS_TYPE = {
    AveragingMode.CYCLIC: {
        AcquisitionType.INTEGRATION: AveragedIntegratedResults,
        AcquisitionType.RAW: AveragedRawWaveformResults,
        AcquisitionType.DISCRIMINATION: AveragedSampleResults,
    },
    AveragingMode.SINGLESHOT: {
        AcquisitionType.INTEGRATION: IntegratedResults,
        AcquisitionType.RAW: RawWaveformResults,
        AcquisitionType.DISCRIMINATION: SampleResults,
    },
}


@dataclass(frozen=True)
class ExecutionParameters:
    """Data structure to deal with execution parameters"""

    nshots: Optional[int] = None
    """Number of shots to sample from the experiment. Default is the runcard value."""
    relaxation_time: Optional[int] = None
    """Time to wait for the qubit to relax to its ground Sample between shots in ns. Default is the runcard value."""
    fast_reset: bool = False
    """Enable or disable fast reset"""
    acquisition_type: AcquisitionType = AcquisitionType.DISCRIMINATION
    """Data acquisition type"""
    averaging_mode: AveragingMode = AveragingMode.SINGLESHOT
    """Data averaging mode"""

    @property
    def results_type(self):
        """Returns corresponding results class"""
        return RESULTS_TYPE[self.averaging_mode][self.acquisition_type]



import re
import sys


class SymbolicExpression:
    class InstancesDict(dict):
        def __init__(self, parent):
            self.parent = parent

        def __setitem__(self, __k, __v):
            self.parent.collect_garbage()
            return super().__setitem__(__k, __v)

    @classmethod
    def initialise(cls):
        cls.count = 0
        cls.instances = SymbolicExpression.InstancesDict(cls)

    @staticmethod
    def clear_instances():
        SymbolicExpression.count = 0
        SymbolicExpression.instances.clear()

    @staticmethod
    def collect_garbage():
        num_items = len(SymbolicExpression.instances)
        n = 0
        while len(SymbolicExpression.instances) > 0:
            symbol = list(SymbolicExpression.instances)[n]
            item = SymbolicExpression.instances[symbol]

            external_ref_count = sys.getrefcount(item)

            internal_ref_count = 0
            for se in list(SymbolicExpression.instances):
                if re.search(rf"\b{re.escape(symbol)}\b", SymbolicExpression.instances[se].expression):
                    internal_ref_count += 1

            if not internal_ref_count and external_ref_count < 4:
                # DEBUG SymbolicExpression Garbage Collector Activity
                # print(f'deleting {symbol} with refcount {external_ref_count}')
                del SymbolicExpression.instances[symbol]
                num_items = len(SymbolicExpression.instances)
                n = 0
            else:
                n += 1
            if n == num_items:
                break

    @staticmethod
    def print_reference_counts():
        for symbol, se in SymbolicExpression.instances.items():
            print(symbol, sys.getrefcount(se))

    class CircularReferenceError(Exception):
        pass

    class InvalidExpressionError(Exception):
        pass

    supported_types = (int, float)

    def __init__(
        self, type: type, expression=0, symbol: str = ""
    ):  # (self, expression: str|{self.type}|SymbolicExpression = 0, symbol: str = ''):
        if type not in self.supported_types:
            raise TypeError(f"type argument should be int or float, got {type.__name__}")
        self.type = type

        self._symbol: str = ""
        self._expression: str = ""

        if symbol == "":
            while True:
                symbol = "_sym_" + type.__name__ + str(SymbolicExpression.count)
                SymbolicExpression.count += 1
                if symbol not in SymbolicExpression.instances.keys():
                    break

        self.expression = expression
        self.symbol = symbol

        ################################ Settable Interface (Quantify)

        self.label: str = ""
        self.unit: str = ""

    def set(self, value):
        self.value = value

    @property
    def name(self):
        return self._symbol

    ################################ Settable Interface (Quantify)

    @property
    def symbol(self) -> str:
        return self._symbol

    @symbol.setter
    def symbol(self, symbol: str):
        if not isinstance(symbol, str):
            raise TypeError(f"symbol argument type should be str, got {type(symbol).__name__}")
        if symbol in SymbolicExpression.instances.keys():
            pass  # Allows overwriting
            # raise KeyError(f"symbol should be unique, there is already a SymbolicExpression with symbol {symbol}: {SymbolicExpression.instances[symbol]}")
        if self._symbol == "":
            # Creation
            SymbolicExpression.instances[symbol] = self
        else:
            # Renaming
            SymbolicExpression.instances[symbol] = self  # Add a new reference with the new symbol
            if not self._symbol == symbol:
                del SymbolicExpression.instances[self._symbol]  # Remove the previous reference

            for se in SymbolicExpression.instances.values():  # Update all SymbolicExpressions with the symbol change
                match_string = rf"\b{re.escape(self._symbol)}\b"
                replacement = symbol
                try:
                    se.expression = re.sub(match_string, replacement, se.expression)
                except (SymbolicExpression.InvalidExpressionError, SymbolicExpression.CircularReferenceError) as e:
                    pass
        self._symbol = symbol
        # test for CircularReferenceError
        self.evaluate(self._expression, self._symbol)

    def __getitem__(self, symbol):
        self.symbol = symbol
        return self

    @property
    def expression(self):
        return self._expression

    @expression.setter
    def expression(self, expression):  # (self, expression: str|{self.type}|SymbolicExpression):
        if isinstance(expression, str):
            # evaluate the str expression to confirm it is valid before assigning it
            self.evaluate(expression)
            self._expression = self._replace_internal_symbols(expression)
        elif isinstance(expression, self.type):
            self._expression = str(expression)
        elif isinstance(expression, SymbolicExpression):
            self._expression = expression._expression
            # self.symbol = expression._symbol #(TODO find a solution so that intermediate operations are not stored as instances)
        else:
            raise TypeError(
                f"expression argument type should be {self.type.__name__} or SymbolicExpression, got {type(expression).__name__}"
            )

    @property
    def value(self):  # -> {self.type}:
        return self.evaluate(self._expression, self._symbol)

    @value.setter
    def value(self, value):  # (self, value: {self.type})
        if isinstance(value, self.type):
            self._expression = str(value)
        else:
            raise TypeError(f"value argument type should be {self.type.__name__}, got {type(value).__name__}")

    @property
    def is_constant(self) -> bool:
        try:
            return str(self.type(self._expression)) == self._expression
        except:
            return False

    def __repr__(self):
        try:
            response = f"{self._symbol}: {self._expression} = {self.value}"
        except SymbolicExpression.CircularReferenceError:
            response = f"{self._symbol}: {self._expression} = CircularReferenceError"
        except SymbolicExpression.InvalidExpressionError:
            response = f"{self._symbol}: {self._expression} = InvalidExpressionError"
        return response

    def evaluate(self, expression: str, *previous_evaluations):  # -> {self.type}:
        for symbol in SymbolicExpression.instances.keys():
            if symbol in expression:
                if symbol in previous_evaluations:
                    raise SymbolicExpression.CircularReferenceError(
                        f"Circular Reference evaluating {expression}, variable {symbol} found in {previous_evaluations}"
                    )
                match_string = rf"\b{re.escape(symbol)}\b"
                replacement = str(
                    self.evaluate(
                        SymbolicExpression.instances[symbol]._expression,
                        *previous_evaluations,
                        SymbolicExpression.instances[symbol]._symbol,
                    )
                )
                expression = re.sub(match_string, replacement, expression)
        try:
            invalid_characters = "abcdefghijklmnopqrstuvwxyz_"
            expression = expression.lower()
            for char in invalid_characters:
                if char in expression:
                    raise Exception()
            result = eval(expression)
        except:
            raise SymbolicExpression.InvalidExpressionError(f"Unable to evaluate expression: {expression}")

        if not isinstance(result, self.type):
            raise SymbolicExpression.InvalidExpressionError(
                f"The evaluation of the expression: {expression} does not return {self.type.__name__}"
            )
        return result

    def _replace_internal_symbols(self, expression: str, *previous_evaluations):  # -> {self.type}:
        symbol: str
        for symbol in SymbolicExpression.instances.keys():
            if symbol.startswith("_sym_") and symbol in expression:
                if symbol in previous_evaluations:
                    raise SymbolicExpression.CircularReferenceError(
                        f"Circular Reference evaluating {expression}, variable {symbol} found in {previous_evaluations}"
                    )
                match_string = rf"\b{re.escape(symbol)}\b"
                replacement = str(
                    self._replace_internal_symbols(
                        SymbolicExpression.instances[symbol]._expression,
                        *previous_evaluations,
                        SymbolicExpression.instances[symbol]._symbol,
                    )
                )
                expression = re.sub(match_string, replacement, expression)

        return expression

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        return str(self.value)

    def __lt__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value < other.value
        if isinstance(other, self.supported_types):
            return self.value < other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __gt__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value > other.value
        if isinstance(other, self.supported_types):
            return self.value > other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __le__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value <= other.value
        if isinstance(other, self.supported_types):
            return self.value <= other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __ge__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value >= other.value
        if isinstance(other, self.supported_types):
            return self.value >= other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __eq__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value == other.value
        if isinstance(other, self.supported_types):
            return self.value == other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __ne__(self, other):
        if isinstance(other, SymbolicExpression):
            return self.value != other.value
        if isinstance(other, self.supported_types):
            return self.value != other
        raise TypeError(
            f"Comparison operators expect SymbolicExpression or {self.type.__name__} arguments, got {type(other).__name__}"
        )

    def __add__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({self.symbol} + {other.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({self.symbol} + {str(other)})")

    def __radd__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({other.symbol} + {self.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({str(other)} + {self.symbol})")

    def __sub__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({self.symbol} - {other.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({self.symbol} - {str(other)})")

    def __rsub__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({other.symbol} - {self.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({str(other)} - {self.symbol})")

    def __mul__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({self.symbol} * {other.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({self.symbol} * {str(other)})")

    def __rmul__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({other.symbol} * {self.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({str(other)} * {self.symbol})")

    def __floordiv__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({self.symbol} // {other.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({self.symbol} // {str(other)})")

    def __rfloordiv__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({other.symbol} // {self.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({str(other)} // {self.symbol})")

    def __mod__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({self.symbol} % {other.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({self.symbol} % {str(other)})")

    def __rmod__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            return self.__class__(f"({other.symbol} % {self.symbol})")
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            return self.__class__(f"({str(other)} % {self.symbol})")

    def __iadd__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            self.expression = f"({self.expression} + {other.symbol})"
            return self
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            if self.is_constant:
                self.expression = str(self.type(self.expression) + other)
            else:
                self.expression = f"({self.expression} + {str(other)})"
            return self

    def __isub__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            self.expression = f"({self.expression} - {other.symbol})"
            return self
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            if self.is_constant:
                self.expression = str(self.type(self.expression) - other)
            else:
                self.expression = f"({self.expression} - {str(other)})"
            return self

    def __imul__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            self.expression = f"({self.expression} * {other.symbol})"
            return self
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            if self.is_constant:
                self.expression = str(self.type(self.expression) * other)
            else:
                self.expression = f"({self.expression} * {str(other)})"
            return self

    def __ifloordiv__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            self.expression = f"({self.expression} // {other.symbol})"
            return self
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            if self.is_constant:
                self.expression = str(self.type(self.expression) // other)
            else:
                self.expression = f"({self.expression} // {str(other)})"
            return self

    def __imod__(self, other):  # -> SymbolicExpression:
        if isinstance(other, SymbolicExpression):
            self.expression = f"({self.expression} % {other.symbol})"
            return self
        if (
            self.type == float
            and isinstance(other, self.supported_types)
            or self.type == int
            and isinstance(other, int)
        ):
            if self.is_constant:
                self.expression = str(self.type(self.expression) % other)
            else:
                self.expression = f"({self.expression} % {str(other)})"
            return self

    def __neg__(self):  # -> SymbolicExpression:
        return self.__class__(f"-{self.symbol}")

    def __hash__(self):
        return hash(self._symbol)


class intSymbolicExpression(SymbolicExpression):
    def __init__(self, expression=0, symbol: str = ""):
        super().__init__(int, expression, symbol)


class floatSymbolicExpression(SymbolicExpression):
    def __init__(self, expression=0, symbol: str = ""):
        super().__init__(float, expression, symbol)


SymbolicExpression.initialise()

# TODO
# t0, t1 = SymbolicExpression(0, 't0', 't1') # t0 = 0, t1 = 0
# t0, t1 = SymbolicExpression([0, 5], ['t0', 't1']) # t0 = 0, t1 = 5
# or even better with a dictionary
# tv_dict = {}
# tv_dict = SymbolicExpression({t0: 0, t1: 5}) # t0 = 0, t1 = 5

# TODO
# add all integer and float operators

# TODO
# make compatible with numpy types int32, float, etc....
